(*

Unidad que implementa un Algoritmo de Colonias de Hormigas para un problema Multiobjetivo,
con 4 objetivos.

En concreto se implementa el llamado CHAC4 (Colonia de Hormigas ACorazadas  de 4 objetivos)

Autor:  Antonio Mora Garca

*)

(*
  Obj1 -> Recursos consumidos
  Obj2 -> Distancia al punto destino
  Obj3 -> Energa consumida
  Obj4 -> Visibilidad de la celda
*)

unit CHAC4;

interface

    uses Tipos, Simulador, Math, Dialogs, SysUtils;

(* ********************* TIPOS DEFINIDOS ********************** *)
type

  // estructura con la Solucin para 4 objetivos (fitness de cada objetivo,
  // consumos de recursos y salud, y camino solucin)
  t_solucion4 = record
    fit_obj1, fit_obj2, fit_obj3, fit_obj4: real;
    fit_obj1_2o, fit_obj2_2o: real;   // fitness para recursos y salud
    consum_rec, consum_sal: real;
    camino: t_camino;
  end;

  // estructura para los datos de las hormigas
  t_hormiga4 = record
     salud: real;
     recursos: real;
     solucion: t_solucion4;
  end;

  // estructura para guardar el Frente de Pareto
  t_frente_pareto4 = record
     num_sols_FP: integer;                  // nmero de solucs que hay en el FP
     soluciones_FP: array of t_solucion4;    // soluciones del FP
  end;

(* ********************* CONSTANTES ********************** *)
const

  // Importancia VO
  KVALOBS: single = 0.2;   // constante de valoracin de observacin (ser localizado por enemigo)

  // vecinos a los que se puede pasar desde una celda cualquiera
  N_VECINOS_FIL: byte = 3;
  N_VECINOS_COL: byte = 3;

(* ********************* VARIABLES GLOBALES ********************** *)
var
(*  // --------------------- Constantes del Programa ----------------------

  // Parmetros del SCH
  ALPHA: byte;    // ponderacin feromona (1)
  BETA: byte;     // ponderacin heurstica (2)
  RHO: single;     // parmetro de evaporacin de feromona (0.1)
  q0: single;     // parmetro de explotacin/exploracin en SCH (0.4)

  PHI: single;    // parmetro de evaporacin local de feromona en SCH monoobjetivo

  // Constantes de Ponderacin de las Heursticas
  // N1 (Camino ms Rpido)
  KN1_REC: single;    // el consumo de recursos es importante (0.1)
  KN1_VO:  single;    // las zonas vistas y ocultas es secundario (0.05)
  KN1_DIS: single;    // la distancia debe irse reduciendo (0.85)

  // N2 (Camino ms Seguro)
  KN2_DAN: single;    // el evitar daos es importante (0.1)
  KN2_VO:  single;    // las zonas vistas y ocultas es muy importante (0.85)
  KN2_DIS: single;    // la distancia al punto final (poco relevante, pero que hay que considerar) (0.05)

  // Constantes de Ponderacin de la Funcin de Evaluacin
  KEV1_ZVO: single;   // Zonas Vistas/Ocultas en objetivo 1: camino ms rpido (0.5)
  KEV2_ZVO: single;   // Zonas Vistas/Ocultas en objetivo 2: camino ms seguro (10)
*)
  // ----------------------------------------------------------------

  // Parmetros del Programa
  N_ITERACIONES: integer;
  N_HORMIGAS: integer;

  // vars para depurar fallos
  dIT,dH: integer;

  // ponderacin de los objetivos (O1=LF, O2=LF, O3=(1-LF), O4=(1-LF))
  LF: single;

  // momento de aplicacin de la ponderacin (al principio o al final)
  APLIC_LAMBDA: string;    // 'aLBD' -> anterior, 'pLBD' -> posterior

  // Constantes del Simulador (de las hormigas)
  SALUD_H: real;      //1000;
  RECURSOS_H: real;   // 2000;

  // distancia de Zona Vista/Oculta
  DVO: integer;       // 2

  // parmetros de la Bsqueda Local
  HACER_BL: boolean;
  N_ITERACS_BL: integer;

  // dimensiones del Tablero de bsqueda (porcin del mapa)
  N_FIL_TAB: integer;
  N_COL_TAB: integer;

  // valores iniciales de feromona
  T01: single;  // feromona inicial objetivo 1 (recursos)
  T02: single;  // feromona inicial objetivo 2 (distanciaObj)
  T03: single;  // feromona inicial objetivo 3 (salud)
  T04: single;  // feromona inicial objetivo 4 (visibilidad)

  // Distancia (eucldea) Mxima en el tablero
  DIST_EU_MAX: real;

  // Tablero de bsqueda del algoritmo
  Tablero: t_tablero;

  // Vector con los datos de las Hormigas
  Hormigas: array of t_hormiga4;

  // Frente de Pareto (conjunto de soluciones no dominadas)
  Frente_Pareto: t_frente_pareto4;

  // Mejor solucin del SCH monoobjetivo
  mejor_sol: t_solucion4;

  // MATRICES ( [Filas][Columnas][3][3] )

  // Matrices de Feromona
  MT1: array of array of array of array of double;
  MT2: array of array of array of array of double;
  MT3: array of array of array of array of double;
  MT4: array of array of array of array of double;

  // Matrices de Heursticas
  MN1: array of array of array of array of double;
  MN2: array of array of array of array of double;
  MN3: array of array of array of array of double;
  MN4: array of array of array of array of double;

  // Matriz de valores de zona oculta media para cada nodo
  ZonasO: array of array of real;

  // Matriz de valores de consumos de salud (daos) para cada nodo
  Danios: array of array of real;

  // Nodos de Inicio y Fin del Camino a buscar
  NODO_INICIAL: t_celda;
  NODO_FINAL: t_celda;

  // variable de seleccin de nodo siguiente
  // ('RT' -> Regla de Transicin, 'RD' -> Ruleta de Dominancia)
  SEL_SIG_NODO: string;

  // determina si el enemigo se conoce o si se busca
  ENEMIGOS_CONOCIDOS: boolean;
  // si no se conoce determina cada cuntos pasos se busca al enemigo
  PASOS_BUSCAR_ENEMIGOS: integer;

  // (si el enemigo es conocido) array de situacin de enemigos y
  // nmero de enemigos conocidos
  NODOS_ENEMIGOS: t_lista_enemigos;
  num_enemigos_conocidos: byte;

  VARIAS_EJECUCIONES: boolean;

(* **************** FUNCIONES Y PROCEDIMIENTOS PBLICOS ***************** *)

Function Principal_CHAC_4objs (fich_mapa: string; vlf: real; aplic_lbd: string; DisZVO: integer; sel_nodo: string;
                               salud_cia: real; recursos_cia: real; num_it: integer;
                               num_horm: integer; busq_local: boolean; iteracs_bl: integer;
                               refinar: boolean; SEM: integer; enem_conocido: boolean; pasos_busq_enems: integer;
                               fich_sols: string; varias_ejecs: boolean; dat_estadisticas: boolean): Integer;


implementation

(******************************************************************************)
(************************* FUNCIONES Y PROCEDIMIENTOS *************************)
(******************************************************************************)

(*
   Procedimiento que reserva memoria para las matrices dinmicas.
   No recibe parmetros puesto que trabaja con variables globales.
*)
procedure reservar_memoria();
begin

  // Tablero de Bsqueda
  SetLength(Tablero,N_FIL_TAB,N_COL_TAB);

  // Vector con los datos de las Hormigas
  SetLength(Hormigas,N_HORMIGAS);

  // Matrices de Feromona
  SetLength(MT1,N_FIL_TAB,N_COL_TAB,N_VECINOS_FIL,N_VECINOS_COL);
  SetLength(MT2,N_FIL_TAB,N_COL_TAB,N_VECINOS_FIL,N_VECINOS_COL);
  SetLength(MT3,N_FIL_TAB,N_COL_TAB,N_VECINOS_FIL,N_VECINOS_COL);
  SetLength(MT4,N_FIL_TAB,N_COL_TAB,N_VECINOS_FIL,N_VECINOS_COL);

  // Matrices de Heursticas
  SetLength(MN1,N_FIL_TAB,N_COL_TAB,N_VECINOS_FIL,N_VECINOS_COL);
  SetLength(MN2,N_FIL_TAB,N_COL_TAB,N_VECINOS_FIL,N_VECINOS_COL);
  SetLength(MN3,N_FIL_TAB,N_COL_TAB,N_VECINOS_FIL,N_VECINOS_COL);
  SetLength(MN4,N_FIL_TAB,N_COL_TAB,N_VECINOS_FIL,N_VECINOS_COL);

  // Frente de Pareto
  SetLength(Frente_Pareto.soluciones_FP,0);
  Frente_Pareto.num_sols_FP:= 0;

  // Zonas Ocultas Medias
  SetLength(ZonasO,N_FIL_TAB,N_COL_TAB);

  // Daos
  SetLength(Danios,N_FIL_TAB,N_COL_TAB);

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que libera la memoria de las matrices dinmicas.
   No recibe parmetros puesto que trabaja con variables globales.
*)
procedure liberar_memoria();
begin

  // Tablero de Bsqueda
  Finalize(Tablero);

  // Vector con los datos de las Hormigas
  Finalize(Hormigas);

  // Matrices de Feromona
  Finalize(MT1);
  Finalize(MT2);
  Finalize(MT3);
  Finalize(MT4);

  // Matrices de Heursticas
  Finalize(MN1);
  Finalize(MN2);
  Finalize(MN3);
  Finalize(MN4);

  // Frente de Pareto
  Finalize(Frente_Pareto.soluciones_FP);

  // Zonas Ocultas Medias
  Finalize(ZonasO);

  // Daos
  Finalize(Danios);

  // Vector de Enemigos
  Finalize(NODOS_ENEMIGOS);

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que lee el fichero con los datos del mapa y rellena
   las matriz de celdas (Tablero) con esos valores
   Recibe como parmetros:
      - nombre_fichero -> fichero con los datos
*)
procedure leer_tablero(nombre_fichero: string);
var
   fich_mapa: File of t_celda;
   celda_aux: t_celda;
   i,j: integer;

begin

  // se abre el fichero para lectura
  AssignFile(fich_mapa, nombre_fichero);
  Reset(fich_mapa);

  // se lee la primera celda (informacin del tamao del mapa/tablero)
  if not Eof(fich_mapa) then
     Read(fich_mapa,celda_aux);

  // Reservamos memoria para las matrices (una vez leemos el tamao que deben tener)
  N_FIL_TAB:= celda_aux.y;
  N_COL_TAB:= celda_aux.x;

  reservar_memoria();

  // Leemos el resto del mapa
  num_enemigos_conocidos:= 0;
  i:= 0;
  j:= 0;
  while not Eof(fich_mapa) do
  begin
     Read(fich_mapa,Tablero[i,j]);

     // comprobamos cuando leemos los nodos origen, fin y enemigo
     if (Tablero[i,j].tipo_esp = PORIGEN) then
        NODO_INICIAL:= Tablero[i,j];

     if (Tablero[i,j].tipo_esp = PDESTINO) then
        NODO_FINAL:= Tablero[i,j];

     // si los enemigos son conocidos se leen del mapa
     if (ENEMIGOS_CONOCIDOS = TRUE) then begin

        // se guardan las celdas con los datos de los enemigos
        if (Tablero[i,j].tipo_esp = ENEMIGO) then begin
           num_enemigos_conocidos:= num_enemigos_conocidos + 1;
           SetLength(NODOS_ENEMIGOS,num_enemigos_conocidos);
           NODOS_ENEMIGOS[num_enemigos_conocidos-1]:= Tablero[i,j];
        end;

     end;

     // pasamos al siguiente elemento
     j:= j+1;
     if (j=N_COL_TAB) then begin
        j:= 0;
        i:= i+1;
        if (i=N_FIL_TAB) then i:= 0;
     end;
  end;

  // cerramos el fichero
  CloseFile(fich_mapa);

end;

//------------------------------------------------------------------------------
//----------------- FUNCIONES DE HEURSTICAS (CHAC4) ---------------------------
//------------------------------------------------------------------------------

(*
   Procedimiento que determina si una celda (nodo) es vista u oculta desde las celdas (nodos)
   que la rodean en un radio determinado y calcula una media de
   vistas y ocultas como funcin heurstica.
   Se mira en un rea hexagonal centrada en el nodoi.
   Recibe como parmetros:
      - celdau -> nodo central de la zona
      - radio_ca -> radio en el que se mirar
      - Tablero -> mapa en el que se dibuja
      - max_x -> lmite de columnas del mapa
      - max_y -> lmite de filas del mapa
   Devuelve:
      - val_V -> media de celdas vistas
      - val_O -> media de celdas ocultas
*)
procedure valor_zona_VO_hex (celdau: t_celda; radio_ca: integer; var Tablero: t_tablero;
                             max_x: integer; max_y: integer; var val_V:real; var val_O: real);
var
   i,j,fil,col,c: integer;

   limNO_x, limNO_y: integer;
   inc_nc, num_cels, VO, cont: integer;

   celdac: t_celda;

   cont_v, cont_o: integer;

begin

  // calculamos las coordenadas del punto extremo NorOeste del rea
  // columna (x)
  limNO_x:= celdau.x - radio_ca;

  // si columna final es PAR
  if ((limNO_x mod 2) = 0) then begin
     // fila (y)
     limNO_y:= celdau.y - (radio_ca div 2);
  end
  else begin
     // Columna origen PAR
     if ((celdau.x mod 2) = 0) then
        // fila (y)
        limNO_y:= celdau.y - ((radio_ca div 2)+1)
     else
        // fila (y)
        limNO_y:= celdau.y - (radio_ca div 2);
  end;


  (** se recorren los nodos del rea de izquierda a derecha y de arriba a abajo **)

  // inicializaciones
  fil:= limNO_y;
  col:= limNO_x;

  inc_nc:= 1;
  num_cels:= radio_ca;

  cont:= 0;
  cont_v:= 0;
  cont_o:= 0;

  if ((col mod 2)=0) then c:= 1
  else c:= 0;

  (* bucle exterior (de izquierda a derecha). De NorOeste a NorEste *)
  for i:= 1 to ((radio_ca*2)+1) do begin

     // nmero de celdas a explorar (de arriba a abajo)
     num_cels:= num_cels + inc_nc;

     (* bucle interior (de arriba a abajo) *)
     for j:=1 to num_cels do begin

        // si se sale de los lmites, no se hace nada
        if ((fil<0) or (fil>=max_y)) or ((col<0) or (col>=max_x)) then begin
           fil:= fil+1;
           continue;
        end;

        celdac:= Tablero[fil,col];

        // si la celda es la misma que la central, no se hace nada
        if (celdac.x = celdau.x) and (celdac.y = celdau.y) then begin
           fil:= fil+1;
           continue;
        end;

        // comprobamos si la celda es vista u oculta
        VO:= celda_VO(celdac,celdau,Tablero,(max_x-1),(max_y-1));

        if (VO = 1) then
            cont_v:= cont_v + 1    // es vista
        else
            cont_o:= cont_o + 1;   // es oculta

        // contador del celdas comprobadas
        cont:= cont + 1;


        // la fila va cambiando (creciendo hacia abajo)
        fil:= fil+1;

     end; // for j


     // se deja la fila como estaba
     fil:= fil-num_cels;

     c:= c+1;

     // columnas a la izquierda de la central
     if (i<radio_ca) then begin
        if (c=2) then begin
          c:= 0;
          fil:= fil-1;  // cada 2 iteraciones se resta 1
        end;
        col:= col+1;
     end
     // columna central
     else if (i=radio_ca) then begin
        if (c=2) then begin
          fil:= fil-1;  // cada 2 iteraciones se resta 1
        end;

        col:= col+1;

        // se inicializa c en funcin de si la celda est en columna PAR o IMPAR
        if ((col mod 2) = 0) then c:= 0
        else c:= 1;

        //fil:= fil+1;
     end
     // columnas a la derecha de la central
     else begin
        if (c=2) then begin
          c:= 0;
          fil:= fil+1;  // cada 2 iteraciones se suma 1
        end;
        col:= col+1;

        inc_nc:= -1;
     end;


  end; // for i

  // calculamos las medias de vistas y ocultas
  val_v:= cont_v/cont;
  val_o:= cont_o/cont;

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que calcula las zonas ocultas medias alrededor de cada celda y
   los guarda en una matriz, para poder aceder posteriormente sin recalcularlas.
*)
procedure calcular_zonas_ocultas_medias();
var
  i,j,k: integer;
  val_v: real;
  VO: integer;
  primero_que_ve: boolean;

begin

  for i:=0 to (N_FIL_TAB-1) do begin

    for j:=0 to (N_COL_TAB-1) do begin

         // si es un obstculo o un enemigo, pasamos al siguiente
         if (Tablero[i,j].tipo = OBST) or (Tablero[i,j].tipo_esp = ENEMIGO) then begin
             ZonasO[i,j]:= 0;
             continue;
         end;

        // si no hay enemigo en el mapa o no se conoce la posicin del enemigo
        // y el radio de la zona a mirar es menor a la capacidad de adquisicin
        // del enemigo (podra ver a la unidad y estar fuera de la zona),
        // se calcula la zona de celdas ocultas medias
        // ZVO ser la capacidad de adquisicin de la Unidad
        if (num_enemigos_conocidos=0) or
           ((ENEMIGOS_CONOCIDOS=FALSE) and (DVO<CAP_ADQ_ENE)) then begin
           valor_zona_VO_hex(Tablero[i,j],DVO,Tablero,N_COL_TAB,N_FIL_TAB,val_v,ZonasO[i,j]);
        end
        // si se conoce (haba en el mapa o se ha encontrado)
        // se calcula la visibilidad de la celda desde cada enemigo
        else if (num_enemigos_conocidos>0) then begin

           ZonasO[i,j]:=1;
           primero_que_ve:= TRUE;

           for k:=0 to (num_enemigos_conocidos-1) do begin

              VO:= celda_VO(Tablero[i,j],NODOS_ENEMIGOS[k],Tablero,(N_COL_TAB-1),(N_FIL_TAB-1));
              if (VO=1) then begin
                 ZonasO[i,j]:= 0;
                 break;
              end;
               (*  // cuando la unidad sea localizada, el valor a eliminar es muy grande
                 if (primero_que_ve=TRUE) then begin
                    ZonasO[i,j]:= KVALOBS;
                    primero_que_ve:= FALSE;
                 end
                 else
                   ZonasO[i,j]:= ZonasO[i,j] - (KVALOBS/(num_enemigos_conocidos-1));

                 end;
              *)
           end; // for k

           (* // se calcula el tanto por 1 de celdas desde la que es oculta
              ZonasO[i,j]:= ZonasO[i,j]/num_enemigos_conocidos;  *)

        end  // else if (enems_conocidos > 0)
        // no hay enemigos conocidos, pero el radio de la zona a mirar es mayor o igual
        // que la capac. de adquisicin del enemigo (ningn enemigo puede ver a
        // la unidad desde fuera de la zona)
        else
           ZonasO[i,j]:= 0;

    end; // for j

  end; // for i

  // ya no necesitamos el vector de enemigos (se libera su memoria)
  Finalize(NODOS_ENEMIGOS);

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que calcula los daos que se produciran en cada celda y
   los guarda en una matriz, para poder aceder posteriormente sin recalcularlas.
*)
procedure calcular_danios();
var
  i,j: integer;

begin

  for i:=0 to (N_FIL_TAB-1) do
    for j:=0 to (N_COL_TAB-1) do
        // el dao solo depende del nodo final, por lo que ponemos 2 veces el mismo
        Danios[i,j]:= consumo_salud(Tablero[i,j],Tablero[i,j],SALUD_H);

end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//                              HEURSTICAS
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

(*
   Funcin que calcula el valor de la heurstica 1 (MINIMOS RECURSOS)
   para el paso de los nodos que determinan los ndices que recibe como parmetro.
   Recibe como parmetros:
      - fil_i -> fila nodo inicial
      - col_i -> columna nodo inicial
      - fil_vi -> fila del vecino de i (nodo al que se pasar)
      - col_vi -> columna del vecino de i (nodo al que se pasar)
   Devuelve:
      - el valor de la heurstica 1 para esos nodos
*)
function calcular_heuristica1 (fil_i: integer; col_i: integer; fil_vi: integer; col_vi: integer):double;

var
  fil_vec, col_vec: integer;  // ndices del nodo vecino

  m1: double;  // miembros de la frmula

begin

  // obtenemos los ndices del vecino

  // Fila
  fil_vec:= fil_i + fil_vi - 1;

  // si se sale de los lmites del tablero, le asignamos el valor mnimo,
  // ya que no ser un nodo a atravesar
  if (fil_vec < 0) or (fil_vec >= N_FIL_TAB) then begin
     Result:= MinDouble;
     exit;
  end;

  // Columna
  col_vec:= col_i + col_vi - 1;

  // si se sale de los lmites del tablero, le asignamos el valor mnimo,
  // ya que no ser un nodo a atravesar
  if (col_vec < 0) or (col_vec >= N_COL_TAB) then begin
     Result:= MinDouble;
     exit;
  end;

  // si los ndices del nodo son iguales que los del vecino (es el mismo nodo)
  // asignamos el valor mnimo (nunca se pasar de un nodo a l mismo) y terminamos
  if (fil_vec = fil_i) and (col_vec = col_i) then begin
     Result:= MinDouble;
     exit;
  end;


  // @@@ No se consideran las celdas vecinas fuera del hexgono
  if (es_vecino_hexa(fil_i,col_i,fil_vec,col_vec,N_FIL_TAB,N_COL_TAB) = FALSE) then begin
     Result:= MinDouble;
     exit;
  end;


  // si es obstculo alguno de ellos, se termina
  if (Tablero[fil_i,col_i].tipo = OBST) or (Tablero[fil_vec,col_vec].tipo = OBST) then begin
     Result:= MinDouble;
     exit;
  end;

  // si es la celda que ocupa un enemigo alguno de ellos, se termina
  if (Tablero[fil_i,col_i].tipo_esp = ENEMIGO) or (Tablero[fil_vec,col_vec].tipo_esp = ENEMIGO) then begin
     Result:= MinDouble;
     exit;
  end;

  // recursos consumidos
  m1:= 1.0/(consumo_recursos(Tablero[fil_i,col_i],Tablero[fil_vec,col_vec],RECURSOS_H)*MAX_RECURSOS);

  // Frmula de la Funcin Heurstica 1
  Result:= m1; //(KN1_REC * m1);

end;

//------------------------------------------------------------------------------

(*
   Funcin que calcula el valor de la heurstica 2 (MNIMA DISTANCIA A OBJETIVO)
   para el paso de los nodos que determinan los ndices que recibe como parmetro.
   Recibe como parmetros:
      - fil_i -> fila nodo inicial
      - col_i -> columna nodo inicial
      - fil_vi -> fila del vecino de i (nodo al que se pasar)
      - col_vi -> columna del vecino de i (nodo al que se pasar)
   Devuelve:
      - el valor de la heurstica 2 para esos nodos
*)
function calcular_heuristica2 (fil_i: integer; col_i: integer; fil_vi: integer; col_vi: integer):double;

var
  fil_vec, col_vec: integer;  // ndices del nodo vecino

  m2: double;  // miembros de la frmula

begin

  // obtenemos los ndices del vecino

  // Fila
  fil_vec:= fil_i + fil_vi - 1;

  // si se sale de los lmites del tablero, le asignamos el valor mnimo,
  // ya que no ser un nodo a atravesar
  if (fil_vec < 0) or (fil_vec >= N_FIL_TAB) then begin
     Result:= MinDouble;
     exit;
  end;

  // Columna
  col_vec:= col_i + col_vi - 1;

  // si se sale de los lmites del tablero, le asignamos el valor mnimo,
  // ya que no ser un nodo a atravesar
  if (col_vec < 0) or (col_vec >= N_COL_TAB) then begin
     Result:= MinDouble;
     exit;
  end;

  // si los ndices del nodo son iguales que los del vecino (es el mismo nodo)
  // asignamos el valor mnimo (nunca se pasar de un nodo a l mismo) y terminamos
  if (fil_vec = fil_i) and (col_vec = col_i) then begin
     Result:= MinDouble;
     exit;
  end;


  // @@@ No se consideran las celdas vecinas fuera del hexgono
  if (es_vecino_hexa(fil_i,col_i,fil_vec,col_vec,N_FIL_TAB,N_COL_TAB) = FALSE) then begin
     Result:= MinDouble;
     exit;
  end;


  // si es obstculo alguno de ellos, se termina
  if (Tablero[fil_i,col_i].tipo = OBST) or (Tablero[fil_vec,col_vec].tipo = OBST) then begin
     Result:= MinDouble;
     exit;
  end;

  // si es la celda que ocupa un enemigo alguno de ellos, se termina
  if (Tablero[fil_i,col_i].tipo_esp = ENEMIGO) or (Tablero[fil_vec,col_vec].tipo_esp = ENEMIGO) then begin
     Result:= MinDouble;
     exit;
  end;


  // Distancia (desde el vecino al nodo final)
  if (fil_vec <> NODO_FINAL.y) or (col_vec <> NODO_FINAL.x) then
     m2:= 1.0/distancia_eu(Tablero[fil_vec,col_vec],NODO_FINAL)
  else
     m2:= 2;   // si estamos en el nodo final, la distancia ser mnima (heurstica mxima)

  // Frmula de la Funcin Heurstica 2
  Result:= m2; //(KN1_DIS * m2);

end;

//------------------------------------------------------------------------------

(*
   Funcin que calcula el valor de la heurstica 3 (MINIMO DAO)
   para el paso de los nodos que determinan los ndices que recibe como parmetro.
   Recibe como parmetros:
      - fil_i -> fila nodo inicial
      - col_i -> columna nodo inicial
      - fil_vi -> fila del vecino de i (nodo al que se pasar)
      - col_vi -> columna del vecino de i (nodo al que se pasar)
   Devuelve:
      - el valor de la heurstica 3 para esos nodos
*)
function calcular_heuristica3 (fil_i: integer; col_i: integer; fil_vi: integer; col_vi: integer):double;

var
  fil_vec, col_vec: integer;  // ndices del nodo vecino

  m1: double;  // miembros de la frmula

begin

  // obtenemos los ndices del vecino

  // Fila
  fil_vec:= fil_i + fil_vi - 1;

  // si se sale de los lmites del tablero, le asignamos el valor mnimo,
  // ya que no ser un nodo a atravesar
  if (fil_vec < 0) or (fil_vec >= N_FIL_TAB) then begin
     Result:= MinDouble;
     exit;
  end;

  // Columna
  col_vec:= col_i + col_vi - 1;

  // si se sale de los lmites del tablero, le asignamos el valor mnimo,
  // ya que no ser un nodo a atravesar
  if (col_vec < 0) or (col_vec >= N_COL_TAB) then begin
     Result:= MinDouble;
     exit;
  end;

  // si los ndices del nodo son iguales que los del vecino (es el mismo nodo)
  // asignamos el valor mnimo (nunca se pasar de un nodo a l mismo) y terminamos
  if (fil_vec = fil_i) and (col_vec = col_i) then begin
     Result:= MinDouble;
     exit;
  end;


  // @@@ No se consideran las celdas vecinas fuera del hexgono
  if (es_vecino_hexa(fil_i,col_i,fil_vec,col_vec,N_FIL_TAB,N_COL_TAB) = FALSE) then begin
     Result:= MinDouble;
     exit;
  end;


  // si es obstculo alguno de ellos, se termina
  if (Tablero[fil_i,col_i].tipo = OBST) or (Tablero[fil_vec,col_vec].tipo = OBST) then begin
     Result:= MinDouble;
     exit;
  end;

  // si es la celda que ocupa un enemigo alguno de ellos, se termina
  if (Tablero[fil_i,col_i].tipo_esp = ENEMIGO) or (Tablero[fil_vec,col_vec].tipo_esp = ENEMIGO) then begin
     Result:= MinDouble;
     exit;
  end;


  // Calculamos los valores Heursticos (miembros de la funcin)

  // Daos (en la celda vecina)
  if (Danios[fil_vec,col_vec]<>0) then
     m1:= 1.0/(Danios[fil_vec,col_vec]*MAX_DANIO)
  else
     m1:= 2;   // si los daos son 0 ponemos como valor uno grande

  // Frmula de la Funcin Heurstica 3 (se aade la distancia para que sea un poco ms dirigida)
  Result:= m1; //(KN2_DAN * m1);

end;

//------------------------------------------------------------------------------

(*
   Funcin que calcula el valor de la heurstica 4 (MINIMA VISIBILIDAD)
   para el paso de los nodos que determinan los ndices que recibe como parmetro.
   Recibe como parmetros:
      - fil_i -> fila nodo inicial
      - col_i -> columna nodo inicial
      - fil_vi -> fila del vecino de i (nodo al que se pasar)
      - col_vi -> columna del vecino de i (nodo al que se pasar)
   Devuelve:
      - el valor de la heurstica 4 para esos nodos
*)
function calcular_heuristica4 (fil_i: integer; col_i: integer; fil_vi: integer; col_vi: integer):double;

var
  fil_vec, col_vec: integer;  // ndices del nodo vecino

  m2: double;  // miembros de la frmula

begin

  // obtenemos los ndices del vecino

  // Fila
  fil_vec:= fil_i + fil_vi - 1;

  // si se sale de los lmites del tablero, le asignamos el valor mnimo,
  // ya que no ser un nodo a atravesar
  if (fil_vec < 0) or (fil_vec >= N_FIL_TAB) then begin
     Result:= MinDouble;
     exit;
  end;

  // Columna
  col_vec:= col_i + col_vi - 1;

  // si se sale de los lmites del tablero, le asignamos el valor mnimo,
  // ya que no ser un nodo a atravesar
  if (col_vec < 0) or (col_vec >= N_COL_TAB) then begin
     Result:= MinDouble;
     exit;
  end;

  // si los ndices del nodo son iguales que los del vecino (es el mismo nodo)
  // asignamos el valor mnimo (nunca se pasar de un nodo a l mismo) y terminamos
  if (fil_vec = fil_i) and (col_vec = col_i) then begin
     Result:= MinDouble;
     exit;
  end;


  // @@@ No se consideran las celdas vecinas fuera del hexgono
  if (es_vecino_hexa(fil_i,col_i,fil_vec,col_vec,N_FIL_TAB,N_COL_TAB) = FALSE) then begin
     Result:= MinDouble;
     exit;
  end;


  // si es obstculo alguno de ellos, se termina
  if (Tablero[fil_i,col_i].tipo = OBST) or (Tablero[fil_vec,col_vec].tipo = OBST) then begin
     Result:= MinDouble;
     exit;
  end;

  // si es la celda que ocupa un enemigo alguno de ellos, se termina
  if (Tablero[fil_i,col_i].tipo_esp = ENEMIGO) or (Tablero[fil_vec,col_vec].tipo_esp = ENEMIGO) then begin
     Result:= MinDouble;
     exit;
  end;


  // Calculamos los valores Heursticos (miembros de la funcin)

  // Media de zona oculta (en el vecino), o simplemente Zona Vista/Oculta
  // (si el enemigo es conocido)
  m2:= ZonasO[fil_vec,col_vec];

  if (m2=0) then m2:=0.1;

  // Frmula de la Funcin Heurstica 4
  Result:= m2; //(KN2_VO * m2);

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que inicializa las matrices de heursticas en base a los
   criterios estimados.
*)
procedure inicializar_heuristicas();
var
  i,j,k,l: integer;

begin

  for i:=0 to (N_FIL_TAB-1) do
    for j:=0 to (N_COL_TAB-1) do
      for k:=0 to (N_VECINOS_FIL-1) do
        for l:=0 to (N_VECINOS_COL-1) do begin
           MN1[i,j,k,l]:= calcular_heuristica1(i,j,k,l);
           MN2[i,j,k,l]:= calcular_heuristica2(i,j,k,l);
           MN3[i,j,k,l]:= calcular_heuristica3(i,j,k,l);
           MN4[i,j,k,l]:= calcular_heuristica4(i,j,k,l);
        end;


end;

//------------------------------------------------------------------------------

(*
   Procedimiento que carga los datos del tablero a considerar para la bsqueda
   y rellena las matrices de Feromona y Heursticas
   Recibe como parmetros:
      - fichero_mapa -> fichero con los datos del mapa a resolver
*)
procedure cargar_inicializar_datos(fichero_mapa: string);

begin

    // leemos los datos del tablero y marcamos los nodos como obstculos
    leer_tablero(fichero_mapa);

    if ((NODO_INICIAL.altura = High(Integer)) or
        (NODO_FINAL.altura = High(Integer))) then begin
       // no se han leido correctamente los puntos de Origen y Destino
       exit;
    end;

    // ## dinamismo ## //
    //// si hay que buscar al enemigo, se busca desde el origen la primera vez
    //if (ENEMIGOS_CONOCIDOS = FALSE) then
    //   num_enemigos_conocidos:= buscar_enemigos_ca_hex(NODO_INICIAL,CAP_ADQ_UNI,NODOS_ENEMIGOS,
    //                                                  Tablero,N_COL_TAB,N_FIL_TAB);

    // creamos las matrices de ZonasO Medias y Danios
    calcular_zonas_ocultas_medias();
    calcular_danios();

    // 4 matrices en Multiobjetivo
    inicializar_heuristicas()

end;

//------------------------------------------------------------------------------
//------------------ FUNCIONES DE ALGORITMO (CHAC4) ----------------------------
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------

(*
   Procedimiento que inicializa las matrices de feromonas con un valor.
   Recibe como parmetros:
      - valT01 -> valor de la feromona inicial para el objetivo 1
      - valT02 -> valor de la feromona inicial para el objetivo 2
      - valT03 -> valor de la feromona inicial para el objetivo 3
      - valT04 -> valor de la feromona inicial para el objetivo 4
*)
procedure inicializar_feromonas(valT01: single; valT02: single;
                                valT03: single; valT04: single);

var
  i,j,k,l: integer;

begin

  for i:=0 to (N_FIL_TAB-1) do
    for j:=0 to (N_COL_TAB-1) do
      for k:=0 to (N_VECINOS_FIL-1) do
        for l:=0 to (N_VECINOS_COL-1) do begin
           MT1[i,j,k,l]:= valT01;
           MT2[i,j,k,l]:= valT02;
           MT3[i,j,k,l]:= valT03;
           MT4[i,j,k,l]:= valT04;
        end;

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que inicializa los valores de las hormigas.
   Recibe como parmetros:
      - val_salud -> valor de la salud de la hormiga
      - val_recursos -> valor de los recursos de la hormiga
*)
procedure inicializar_hormigas(val_salud: real; val_recursos: real);

var
  i: integer;

begin

    for i:=0 to (N_HORMIGAS-1) do begin
       Hormigas[i].salud:= val_salud;
       Hormigas[i].recursos:= val_recursos;

       // limpiamos la solucin
       Hormigas[i].solucion.fit_obj1:= 0;
       Hormigas[i].solucion.fit_obj2:= 0;
       Hormigas[i].solucion.fit_obj3:= 0;
       Hormigas[i].solucion.fit_obj4:= 0;
       Hormigas[i].solucion.fit_obj1_2o:= 0;
       Hormigas[i].solucion.fit_obj2_2o:= 0;
       Finalize(Hormigas[i].solucion.camino);
    end;

end;

//------------------------------------------------------------------------------

(*
   Funcin que comprueba si dos soluciones son iguales, es decir, si el camino
   lo forman los mismos nodos, en el mismo orden y tienen los mismos fitness para
   ambos objetivos.
   Recibe como parmetros:
      - sol1, sol2 -> soluciones a comparar
   Devuelve como resultado:
      - TRUE -> si las soluciones son iguales
      - FALSE -> si las soluciones no son iguales
*)
function solucs_iguales(sol1: t_solucion4; sol2: t_solucion4):boolean;

var
   i: integer;
   iguales: boolean;

begin

   // si alguno de los fitness es diferente, ya no son iguales
   if (sol1.fit_obj1 <> sol2.fit_obj1) or
      (sol1.fit_obj2 <> sol2.fit_obj2) or
      (sol1.fit_obj3 <> sol2.fit_obj3) or
      (sol1.fit_obj4 <> sol2.fit_obj4) then begin
        Result:= FALSE;
        exit;
   end
   // si tienen distinto nmero de nodos, ya no son iguales
   else if (Length(sol1.camino) <> Length(sol2.camino)) then begin
        Result:= FALSE;
        exit;
   end
   // si no se cumple lo anterior, comparamos los nodos
   else begin

      iguales:= TRUE;

      for i:=low(sol1.camino) to high(sol1.camino) do begin

         // si algn nodo es diferente, ya no sern iguales
         if (sol1.camino[i].y <> sol2.camino[i].y) or
            (sol1.camino[i].x <> sol2.camino[i].x) then begin
             iguales:= FALSE;
             break;
         end;

      end;

   end;

   Result:= iguales;

end;

//------------------------------------------------------------------------------

(*
   Funcin que busca un nodo en una lista.
   Recibe como parmetros:
      - lista_camino -> lista de nodos del camino actual
      - num_nodos_camino -> nmero de nodos en el camino actual
   Devuelve:
      - la posicin del nodo en la lista si est o -1 si no est
*)
function nodo_en_lista(nodo: t_celda; lista_nodos: t_camino; num_nodos: integer): integer;

var
  i: integer;
  pos: integer;

begin

   pos:= -1;

   // recorremos la lista buscando el nodo
   for i:=0 to (num_nodos-1) do begin

      // si est, termina el bucle
      if (lista_nodos[i].y = nodo.y) and (lista_nodos[i].x = nodo.x) then begin
         pos:= i;
         break;
      end;

   end;

   Result:= pos;

end;

//------------------------------------------------------------------------------

(*
   Funcin que construye la lista de nodos alcanzables desde uno dado, es decir, aquellos
   que que no estn ya en el camino, que no sean un obstculo y tales que, al pasar a ellos,
   el consumo de recursos y/o salud no supere los que le restan a la hormiga.
   Recibe como parmetros:
      - nodo -> nodo actual del camino
      - salud -> salud actual de la hormiga
      - recursos -> recursos actuales de la hormiga
      - comp_en_camino -> var que indica si se comprobar si un nodo est ya en el camino o no
      - lista_camino -> lista de nodos del camino actual
      - num_nodos_camino -> nmero de nodos en el camino actual
   Salidas
      - lista_nodos -> lista de nodos alcanzables
   Devuelve
      - nmero de nodos de la lista de candidatos
*)
function lista_alcanzables (nodo: t_celda; salud: real; recursos: real; comp_en_camino: boolean;
                            lista_camino: t_camino; num_nodos_camino: integer;
                            var lista_nodos: t_alcanzables): integer;

var
  i,j: integer;
  fil_c,col_c: integer;

  pos_nodo: integer;

  num_nodos: integer;

  salud_cons,recurs_cons: real;

begin

   // la lista est vaca en principio
   num_nodos:= 0;
   SetLength(lista_nodos, 0);

   (* comprobamos para cada uno de los vecinos de la celda, si es alcanzable o no *)
   for i:=0 to (N_VECINOS_FIL-1) do begin

     // si el vecino est fuera del mapa, pasamos al siguiente
     if ((nodo.y + (i-1)) < 0) or ((nodo.y + (i-1)) >= N_FIL_TAB) then continue;

     for j:=0 to (N_VECINOS_COL-1) do begin

         // si el vecino est fuera del mapa, pasamos al siguiente
         if ((nodo.x + (j-1)) < 0) or ((nodo.x + (j-1)) >= N_COL_TAB) then continue;

         // coordenadas de la celda a comprobar (vecino) (y=fila,x=columna)
         fil_c:= (nodo.y + (i-1));
         col_c:= (nodo.x + (j-1));

         // si fila y columna coinciden, es que es el mismo nodo y no se evala
         if (fil_c = nodo.y) and (col_c = nodo.x) then begin
             continue;
         end;


         // @@@ No se consideran las celdas vecinas fuera del hexgono
         if (es_vecino_hexa(nodo.y,nodo.x,fil_c,col_c,N_FIL_TAB,N_COL_TAB) = FALSE) then begin
            continue;
         end;


         (* *** Comprobamos OBSTCULO *** *)

         // si es un obstculo artificial (definido en el tipo de celda) o
         // si la diferencia de alturas constituye un obstculo natural,
         // pasamos al siguiente
         if (hay_obstaculo(nodo,Tablero[fil_c,col_c]) = TRUE) then begin
            continue;
         end;


         (* *** Comprobamos ENEMIGO *** *)

         // si es la celda de un enemigo, pasamos al siguiente
         if (Tablero[fil_c,col_c].tipo_esp = ENEMIGO) then begin
             continue;
         end;


         // (si hay que comprobarlo) si el nodo ya est en el camino, no se evala
         if (comp_en_camino=TRUE) then begin

            pos_nodo:= nodo_en_lista(Tablero[fil_c,col_c],lista_camino,num_nodos_camino);
            if (pos_nodo <> -1) then begin
              continue;
            end;

         end;

         (* *** Comprobamos REQUISITOS SALUD *** *)

         // si el paso de la celda actual al vecino implica perder ms salud de la que tenemos,
         // descartamos ese vecino (pasamos al mirar el siguiente)
         salud_cons:= consumo_salud (nodo,Tablero[fil_c,col_c],salud);

         if (salud <= salud_cons) then begin
             continue;
         end;

         (* *** Comprobamos REQUISITOS RECURSOS *** *)

         // si el paso de la celda actual al vecino implica un consumo de recursos mayor de los
         // que tenemos, descartamos ese vecino (pasamos al mirar el siguiente)
         recurs_cons:= consumo_recursos (nodo,Tablero[fil_c,col_c],recursos);

         if (recursos <= recurs_cons) then begin
             continue;
         end;


         // si se llega a este punto, es que el nodo es alcanzable y se guarda en la lista
         num_nodos:= num_nodos+1;
         SetLength(lista_nodos,num_nodos);

         lista_nodos[num_nodos-1]:= Tablero[fil_c,col_c];

     end;

   end;

   // Nmero de nodos en la lista
   Result:= num_nodos;

end;


//------------------------------------------------------------------------------

(*
   Funcin que Elige el siguiente nodo a pasar en el camino de la hormiga usando como regla de
   transicin una asignacin de probabilidad en base a la dominancia (se cuenta cuantos nodos
   domina cada uno de los vecinos) y una ruleta de seleccin.
   Recibe como parmetros:
      - nodo -> nodo actual del camino
      - lista_nodos_alc -> lista de nodos alcanzables desde el actual (a los que se puede pasar)
      - num_nodos -> nmero de nodos en la lista de alcanzables
   Devuelve:
      - nodo al que se pasar en base a la regla de transicin
*)
function elegir_siguiente_nodo_ruleta_dominancia (nodo: t_celda; lista_nodos_alc: array of t_celda;
                                                  num_nodos: integer):t_celda;

var
   i,j: byte;

   //ndices auxiliares
   fil_n,col_n,fil_vi,col_vi,fil_vj,col_vj: integer;

   // probabilidades asociadas a los nodos alcanzables (para ruleta de seleccin)
   prob_nodos: array of double;
   prob_rul: double;
   q: real;
   max_prob: double;
   cont_dominados: byte;

   // trminos para ver dominancia
   t1i,t1j,t2i,t2j,t3i,t3j,t4i,t4j: double;

   ind_nodo: integer;

   // suma de probabilidades para la ruleta
   suma_prob, suma_tot_prob: double;

begin

   suma_tot_prob:= 0;

   // reservamos espacio para el vector de probabilidades
   SetLength(prob_nodos,num_nodos);

   // fila,columna del nodo actual en el tablero (y=fila, x=columna)
   fil_n:= nodo.y;
   col_n:= nodo.x;

   (* *** comprobamos los vecinos para ver si alguno destaca sobre los otros *** *)

   for i:= 0 to (num_nodos-1) do begin

      cont_dominados:= 1;  // se inicializa a 1 para que todos tengan probabilidad

      // ndices en los que mirar en las matrices

      // fila,columna en la lista de vecinos (ltimas dimensiones de la matriz de Feromonas y Heursticas)
      // vecino i
      fil_vi:= (lista_nodos_alc[i].y - nodo.y) + 1;   // fila del vecino en la lista de vecinos (+1 para situarlo entre (0,0)-(2,2))
      col_vi:= (lista_nodos_alc[i].x - nodo.x) + 1;   // columna del vecino en la lista de vecinos (+1 para situarlo entre (0,0)-(2,2))

      // terminos del nodo i
      t1i:= ( power(MT1[fil_n,col_n,fil_vi,col_vi],(ALPHA)) *
              power(MN1[fil_n,col_n,fil_vi,col_vi],(BETA)) );

      t2i:= ( power(MT2[fil_n,col_n,fil_vi,col_vi],(ALPHA)) *
              power(MN2[fil_n,col_n,fil_vi,col_vi],(BETA)) );

      t3i:= ( power(MT3[fil_n,col_n,fil_vi,col_vi],(ALPHA)) *
              power(MN3[fil_n,col_n,fil_vi,col_vi],(BETA)) );

      t4i:= ( power(MT4[fil_n,col_n,fil_vi,col_vi],(ALPHA)) *
              power(MN4[fil_n,col_n,fil_vi,col_vi],(BETA)) );


      for j:= 0 to (num_nodos-1) do begin

         if (i=j) then continue;  // un nodo no se comprueba con l mismo

         // ndices en los que mirar en las matrices

         // fila,columna en la lista de vecinos (ltimas dimensiones de la matriz de Feromonas y Heursticas)
         // vecino j
         fil_vj:= (lista_nodos_alc[j].y - nodo.y) + 1;   // fila del vecino en la lista de vecinos (+1 para situarlo entre (0,0)-(2,2))
         col_vj:= (lista_nodos_alc[j].x - nodo.x) + 1;   // columna del vecino en la lista de vecinos (+1 para situarlo entre (0,0)-(2,2))

         // terminos del nodo j (sin aplicar lambda)
         t1j:= ( power(MT1[fil_n,col_n,fil_vj,col_vj],ALPHA) *
                 power(MN1[fil_n,col_n,fil_vj,col_vj],BETA) );

         t2j:= ( power(MT2[fil_n,col_n,fil_vj,col_vj],ALPHA) *
                 power(MN2[fil_n,col_n,fil_vj,col_vj],BETA) );

         t3j:= ( power(MT3[fil_n,col_n,fil_vj,col_vj],ALPHA) *
                 power(MN3[fil_n,col_n,fil_vj,col_vj],BETA) );

         t4j:= ( power(MT4[fil_n,col_n,fil_vj,col_vj],ALPHA) *
                 power(MN4[fil_n,col_n,fil_vj,col_vj],BETA) );

         (* comprobamos si i domina a j e incrementamos un contador de dominados *)
         // el trmino tendr un valor ms alto para un mejor nodo
         if ((t1i > t1j) and (t2i >= t2j) and (t3i >= t3j) and (t4i >= t4j)) or
            ((t1i >= t1j) and (t2i > t2j) and (t3i >= t3j) and (t4i >= t4j)) or
            ((t1i >= t1j) and (t2i >= t2j) and (t3i > t3j) and (t4i >= t4j)) or
            ((t1i >= t1j) and (t2i >= t2j) and (t3i >= t3j) and (t4i > t4j)) then begin

             cont_dominados:= cont_dominados + 1;

         end;

      end;   // for j

      // probabilidad en base a dominancia
      prob_nodos[i]:= cont_dominados;
      suma_tot_prob:= suma_tot_prob + cont_dominados;

   end;   // for i


   (* *** Elegimos el nodo siguiente o en base a una ruleta de probabilidades o el mejor *** *)

   q:= Random;  // nmero aleatorio en [0..1]

   // elegimos el siguiente nodo segn el parmetro q0 del SCH

   if (q<=q0) then begin
      // elegimos el nodo con la mayor probabilidad asociada
      max_prob:= 0;
      ind_nodo:= 0;

      for i:= 0 to (num_nodos-1) do begin

          if (max_prob < prob_nodos[i]) then begin
              max_prob:= prob_nodos[i];
              ind_nodo:= i;
          end;

      end;

   end
   else begin

      // si ninguno domina a otro (se ha sumado 1 en cada nodo), se elige uno de forma aleatoria
      if (suma_tot_prob=(num_nodos-1)) then begin
         ind_nodo:= Random((num_nodos-1));
      end
      else begin

         // lanzamos la ruleta para seleccionar el nodo al que pasar

         // vamos comprobando qu nodo ser elegido en base a la probabilidad
         // asociada simulando una ruleta

         // tirada de la ruleta (no debe salir 0)
         repeat
            prob_rul:= Random;   // nmero entre [0..1]
         until (prob_rul>0);

         // comprobamos el nodo elegido (en base a la tirada)
         ind_nodo:= -1;
         suma_prob:= 0;

         for i:= 0 to (num_nodos-1) do begin
            suma_prob:= suma_prob + (prob_nodos[i]/suma_tot_prob);

            if (suma_prob >= prob_rul) then begin
                ind_nodo:= i;
                break;
            end;

         end;

      end; // else (si alguno domina a otro)


   end;  // else  (si se elige el nodo con ruleta)


   // limpiamos las variables dinmicas
   Finalize(prob_nodos);


   (* devolvemos el nodo elegido *)
   if (ind_nodo=-1) then begin
      ShowMessage('ERROR: There is no Next node in the list');
      exit;
   end;

   Result:= lista_nodos_alc[ind_nodo];

end;

//------------------------------------------------------------------------------

(*
   Funcin que Elige el siguiente nodo a pasar en el camino de la hormiga en base
   a la regla de transicin y una ruleta de seleccin con probabilidades
   Recibe como parmetros:
      - nodo -> nodo actual del camino
      - lista_nodos_alc -> lista de nodos alcanzables desde el actual (a los que se puede pasar)
      - num_nodos -> nmero de nodos en la lista de alcanzables
   Devuelve:
      - nodo al que se pasar en base a la regla de transicin
*)
function elegir_siguiente_nodo_regla_transicion (nodo: t_celda; lista_nodos_alc: array of t_celda;
                                                 num_nodos: integer):t_celda;

var
   i,j: byte;

   //ndices auxiliares
   fil_n,col_n,fil_v,col_v: integer;

   // dividendo y divisor de la frmula de la regla de transicin
   t1: array of array of double;
   t2: double;

   // probabilidades asociadas a los nodos alcanzables (para ruleta de seleccin)
   prob_nodos: array of double;
   prob_rul: double;
   q: real;
   max_prob: double;
   ind_nodo: integer;

   // suma de probabilidades para la ruleta
   suma_prob, suma_tot_prob: double;

begin

   // reservamos espacio para el vector de probabilidades
   SetLength(prob_nodos,num_nodos);

   // reservamos memoria para los valores de cada uno de los miembros de la frmula
   SetLength(t1,N_VECINOS_FIL,N_VECINOS_COL);

   // inicializamos la matriz (por si hay menos candidatos que vecinos)
   for i:=0 to (N_VECINOS_FIL-1) do
     for j:=0 to (N_VECINOS_COL-1) do
        t1[i,j]:= 0;


   // fila,columna del nodo actual en el tablero (y=fila, x=columna)
      fil_n:= nodo.y;
      col_n:= nodo.x;

   // calculamos la probabilidad de cada nodo en base a la REGLA DE TRANSICIN

   (* *** clculo de miembros de la frmula *** *)
   t2:= 0;

   for i:= 0 to (num_nodos-1) do begin

      // ndices en los que mirar en las matrices

      // fila,columna en la lista de vecinos (ltimas dimensiones de la matriz de Feromonas y Heursticas)
      fil_v:= (lista_nodos_alc[i].y - nodo.y) + 1;   // fila del vecino en la lista de vecinos (+1 para situarlo entre (0,0)-(2,2))
      col_v:= (lista_nodos_alc[i].x - nodo.x) + 1;   // columna del vecino en la lista de vecinos (+1 para situarlo entre (0,0)-(2,2))

      (* miembros del dividendo *)

      // aplicamos lambda en este momento o no lo aplicamos (para hacerlo al final)
      if (APLIC_LAMBDA = 'aLBD') then begin
          // guardamos el valor de cada miembro de la sumatoria
          t1[fil_v,col_v]:= ( power(MT1[fil_n,col_n,fil_v,col_v],(LF*ALPHA)) *
                              power(MT2[fil_n,col_n,fil_v,col_v],(LF*ALPHA)) *
                              power(MT3[fil_n,col_n,fil_v,col_v],((1-LF)*ALPHA)) *
                              power(MT4[fil_n,col_n,fil_v,col_v],((1-LF)*ALPHA)) *
                              power(MN1[fil_n,col_n,fil_v,col_v],(LF*BETA)) *
                              power(MN2[fil_n,col_n,fil_v,col_v],(LF*BETA)) *
                              power(MN3[fil_n,col_n,fil_v,col_v],((1-LF)*BETA)) *
                              power(MN4[fil_n,col_n,fil_v,col_v],((1-LF)*BETA)) );
      end
      else begin
          // guardamos el valor de cada miembro de la sumatoria
          t1[fil_v,col_v]:= ( power(MT1[fil_n,col_n,fil_v,col_v],ALPHA) *
                              power(MT2[fil_n,col_n,fil_v,col_v],ALPHA) *
                              power(MT3[fil_n,col_n,fil_v,col_v],ALPHA) *
                              power(MT4[fil_n,col_n,fil_v,col_v],ALPHA) *
                              power(MN1[fil_n,col_n,fil_v,col_v],BETA) *
                              power(MN2[fil_n,col_n,fil_v,col_v],BETA) *
                              power(MN3[fil_n,col_n,fil_v,col_v],BETA) *
                              power(MN4[fil_n,col_n,fil_v,col_v],BETA) );
      end;

      (* clculo del divisor *)
      t2 := t2 + t1[fil_v,col_v];

   end;  // for i

   (* Clculo de las probabilidades *)
   suma_tot_prob:= 0;

   for i:= 0 to (num_nodos-1) do begin

      // ndices en los que mirar en las matrices

      // fila,columna en la lista de vecinos (ltimas dimensiones de la matriz de Feromonas y Heursticas)
      fil_v:= (lista_nodos_alc[i].y - nodo.y) + 1;   // fila del vecino en la lista de vecinos (+1 para situarlo entre (0,0)-(2,2))
      col_v:= (lista_nodos_alc[i].x - nodo.x) + 1;   // columna del vecino en la lista de vecinos (+1 para situarlo entre (0,0)-(2,2))

      (* clculo de la probabilidad asociada a cada transicin de nodos *)

      // probabilidad
      prob_nodos[i]:= (t1[fil_v,col_v])/t2;

      // suma total de las probabilidades (para la ruleta de seleccin)
      suma_tot_prob:= suma_tot_prob + prob_nodos[i];

   end;  // for i


   (* *** Elegimos el nodo siguiente o en base a una ruleta de probabilidades o el mejor *** *)

   q:= Random;  // nmero aleatorio en [0..1]

   // elegimos el siguiente nodo segn el parmetro q0 del SCH

   if (q<=q0) then begin
      // elegimos el nodo con la mayor probabilidad asociada
      max_prob:= 0;
      ind_nodo:= -1;

      for i:= 0 to (num_nodos-1) do begin

          if (max_prob < prob_nodos[i]) then begin
              max_prob:= prob_nodos[i];
              ind_nodo:= i;
          end;

      end;

   end
   else begin

      // lanzamos la ruleta para seleccionar el nodo al que pasar

      // vamos comprobando qu nodo ser elegido en base a la probabilidad
      // asociada simulando una ruleta

      // tirada de la ruleta (no debe salir 0)
      repeat
         prob_rul:= Random;   // nmero entre [0..1]
      until (prob_rul>0);

      // comprobamos el nodo elegido (en base a la tirada)
      ind_nodo:= -1;
      suma_prob:= 0;

      for i:= 0 to (num_nodos-1) do begin
          suma_prob:= suma_prob + (prob_nodos[i]/suma_tot_prob);

          if (suma_prob >= prob_rul) then begin
             ind_nodo:= i;
             break;
          end;

      end;

   end;  // else

   // limpiamos las variables dinmicas
   Finalize(prob_nodos);


   (* devolvemos el nodo elegido *)
   if (ind_nodo=-1) then begin
      ShowMessage('ERROR: There is no Next node in the list');
      exit;
   end;

   Result:= lista_nodos_alc[ind_nodo];

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que realiza la evaporacin de feromona (para penalizar) los arcos
   de una solucin no completa vlida.
   Recibe como parmetros:
      - nodo_act -> nodo actual del camino solucin
      - nodo_sig -> nodo siguiente en el camino solucin
      - tpu_remanente -> tanto por uno de feromona que quedar tras la evaporacin
                         (1 - tanto por uno de evaporacin (penalizacin))
*)
procedure evaporacion_local_feromona(nodo_act: t_celda; nodo_sig:t_celda; tpu_remanente: real);

var
   fil_n,col_n,fil_v,col_v: integer; // ndices de las matrices de feromonas

begin

      // ndices en la matriz de feromonas (nodo actual)
      fil_n:= nodo_act.y;
      col_n:= nodo_act.x;

      // ndices del vecino (nodo siguiente) respecto al anterior
      fil_v:= (nodo_sig.y - fil_n) + 1;   // +1 para situar en [0..2]
      col_v:= (nodo_sig.x - col_n) + 1;   // +1 para situar en [0..2]

      //                                         evaporacin
      MT1[fil_n,col_n,fil_v,col_v]:= (tpu_remanente * MT1[fil_n,col_n,fil_v,col_v]);
      MT2[fil_n,col_n,fil_v,col_v]:= (tpu_remanente * MT2[fil_n,col_n,fil_v,col_v]);
      MT3[fil_n,col_n,fil_v,col_v]:= (tpu_remanente * MT3[fil_n,col_n,fil_v,col_v]);
      MT4[fil_n,col_n,fil_v,col_v]:= (tpu_remanente * MT4[fil_n,col_n,fil_v,col_v]);

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que realiza la evaporacin y aporte de feromona a nivel local
   (en lnea), esto es, cuando se elige el siguiente nodo del camino solucin.
   Recibe como parmetros:
      - nodo_act -> nodo actual del camino solucin
      - nodo_sig -> nodo siguiente en el camino solucin
*)
procedure actualizacion_local_feromona(nodo_act: t_celda; nodo_sig:t_celda);

var
   fil_n,col_n,fil_v,col_v: integer; // ndices de las matrices de feromonas

begin

      // ndices en la matriz de feromonas (nodo actual)
      fil_n:= nodo_act.y;
      col_n:= nodo_act.x;

      // ndices del vecino (nodo siguiente) respecto al anterior
      fil_v:= (nodo_sig.y - fil_n) + 1;   // +1 para situar en [0..2]
      col_v:= (nodo_sig.x - col_n) + 1;   // +1 para situar en [0..2]

      //                                         evaporacin                   aporte
      MT1[fil_n,col_n,fil_v,col_v]:= ((1-RHO) * MT1[fil_n,col_n,fil_v,col_v]) + (RHO*T01);
      MT2[fil_n,col_n,fil_v,col_v]:= ((1-RHO) * MT2[fil_n,col_n,fil_v,col_v]) + (RHO*T02);
      MT3[fil_n,col_n,fil_v,col_v]:= ((1-RHO) * MT3[fil_n,col_n,fil_v,col_v]) + (RHO*T03);
      MT4[fil_n,col_n,fil_v,col_v]:= ((1-RHO) * MT4[fil_n,col_n,fil_v,col_v]) + (RHO*T04);

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que realiza la evaporacin y aporte de feromona a nivel global
   (fuera de lnea), esto es, la que hacen las mejores soluciones obtenidas
   (las que hay en el Frente de Pareto)
*)
procedure actualizacion_global_feromona();

var
  i,j: integer;         // contadores
  fil_n,col_n,fil_v,col_v: integer; // ndices de las matrices de feromonas
  f1,f2,f3,f4: real;    // fitness de los objetivos
  nodo_ant: t_celda;    // nodo anterior en la solucin

begin

   // Se actualiza la feromona de los nodos del camino de cada solucin en el FP
   for i:=low(Frente_Pareto.soluciones_FP) to high(Frente_Pareto.soluciones_FP) do begin

      // SOLUCIN i

      // obtenemos el fitness de cada objetivo de la solucin
      f1:=Frente_Pareto.soluciones_FP[i].fit_obj1;
      f2:=Frente_Pareto.soluciones_FP[i].fit_obj2;
      f3:=Frente_Pareto.soluciones_FP[i].fit_obj3;
      f4:=Frente_Pareto.soluciones_FP[i].fit_obj4;

      if (f4=0) then f4:= 0.1;

      { [*AJUSTE PARAMS*]}
      // se elimina la ponderacin de las constantes para equiparar los aportes
      //f1:= f1/2 + f1/(2*KEV1_ZVO);  // la mitad permanece y en la otra mitad se elimina la ponderacin
      //f2:= f2/2 + f2/(2*KEV2_ZVO);  // la mitad permanece y en la otra mitad se elimina la ponderacin

      // cogemos el nodo inicial
      j:= low(Frente_Pareto.soluciones_FP[i].camino);
      nodo_ant:= Frente_Pareto.soluciones_FP[i].camino[j];

      // recorremos los arcos de la solucin
      for j:=(low(Frente_Pareto.soluciones_FP[i].camino)+1) to
             high(Frente_Pareto.soluciones_FP[i].camino) do begin

          // ndices en la matriz de feromonas
          fil_n:= nodo_ant.y;
          col_n:= nodo_ant.x;

          // ndices del vecino (nodo actual) respecto al anterior
          fil_v:= (Frente_Pareto.soluciones_FP[i].camino[j].y - fil_n) + 1;   // +1 para situar en [0..2]
          col_v:= (Frente_Pareto.soluciones_FP[i].camino[j].x - col_n) + 1;   // +1 para situar en [0..2]

          //                                         evaporacin                   aporte
          MT1[fil_n,col_n,fil_v,col_v]:= ((1-RHO) * MT1[fil_n,col_n,fil_v,col_v]) + (RHO/f1);
          MT2[fil_n,col_n,fil_v,col_v]:= ((1-RHO) * MT2[fil_n,col_n,fil_v,col_v]) + (RHO/f2);
          MT3[fil_n,col_n,fil_v,col_v]:= ((1-RHO) * MT3[fil_n,col_n,fil_v,col_v]) + (RHO/f3);
          MT4[fil_n,col_n,fil_v,col_v]:= ((1-RHO) * MT4[fil_n,col_n,fil_v,col_v]) + (RHO/f4);

          // el nodo actual pasa a ser el anterior
          nodo_ant:= Frente_Pareto.soluciones_FP[i].camino[j];

      end;

   end;


end;

//------------------------------------------------------------------------------

(*
   Procedimiento que Construye una solucin para una hormiga.
   Recibe como parmetros:
      - h -> hormiga que contendr la solucin (camino)
      - nodo_ini -> nodo del que partir la hormiga
      - nodo_fin -> nodo al que se pretende llegar
   Devuelve como resultado:
      - TRUE -> si la solucin es vlida (camino completo y CIA con salud/recursos)
      - FALSE -> si la solucin no es vlida (faltan salud/recursos)

*)
function construir_solucion(var h: t_hormiga4; nodo_ini: t_celda; nodo_fin: t_celda): boolean;

var
  nodos_cam: integer;   // contador de nodos en el camino actual
  nodo_act: t_celda;    // nodo en el que nos encontramos en el camino (nodo actual)
  nodo_sig: t_celda;    // nodo al que se va a pasar (nodo siguiente)

  i, num_nodos: integer;

  // condiciones de fin de construccin de solucin y camino completo (vlido)
  fin_camino: boolean;
  camino_completo: boolean;

  l_nodos_alc: t_alcanzables;  // lista de vecinos alcanzables (en principio 8 mximo)

begin

   (* ponemos el nodo inicial en la lista de nodos que sern el camino *)
   nodos_cam:= 0;
   SetLength(h.solucion.camino,nodos_cam+1);
   h.solucion.camino[nodos_cam]:= nodo_ini;
   nodos_cam:= nodos_cam + 1;

   (* nodo actual *)
   nodo_act:= nodo_ini;
   fin_camino:= FALSE;

   repeat

      (* creamos la lista de nodos alcanzables desde el actual *)
      num_nodos:= lista_alcanzables(nodo_act,h.salud,h.recursos,TRUE,h.solucion.camino,nodos_cam,l_nodos_alc);

      if (num_nodos>0) then begin

         // se elige el siguiente nodo en base a la Regla de Transicin
         if (SEL_SIG_NODO = 'RT') then
            nodo_sig:= elegir_siguiente_nodo_regla_transicion(nodo_act,l_nodos_alc,num_nodos)
         // o en base a la Ruleta de Dominancia
         else
            nodo_sig:= elegir_siguiente_nodo_ruleta_dominancia(nodo_act,l_nodos_alc,num_nodos);

         (* aadimos el nuevo nodo al camino *)
         SetLength(h.solucion.camino,nodos_cam+1);
         h.solucion.camino[nodos_cam]:=nodo_sig;
         nodos_cam:= nodos_cam + 1;

         (* actualizamos la salud y recursos de la hormiga tras pasar a ese nodo *)

         // actualizamos la salud
         h.salud:= h.salud - consumo_salud(nodo_act,nodo_sig,h.salud);

         // actualizamos los recursos
         h.recursos:= h.recursos - consumo_recursos(nodo_act,nodo_sig,h.recursos);


         (* *** EVAPORACIN y APORTE de Feromona al arco recorrido *** *)
         actualizacion_local_feromona(nodo_act,nodo_sig);

         // pasamos al siguiente nodo
         nodo_act:= nodo_sig;

         // si es el ltimo, el camino est terminado y es vlido
         if (nodo_act.x=nodo_fin.x) and (nodo_act.y=nodo_fin.y) then begin
            fin_camino:= TRUE;
            camino_completo:= TRUE;
         end;

      end
      else begin
         (* si ya no hay ms nodos alcanzables, se termina la construccin del camino *)

         // fin de la construccin del camino. No es un camino vlido
         fin_camino:= TRUE;
         camino_completo:= FALSE;

         // se evapora la feromona de los nodos de este camino (no es deseable)
         for i:=0 to (nodos_cam-2) do begin
            evaporacion_local_feromona(h.solucion.camino[i],h.solucion.camino[i+1],(1-(9*RHO)));
         end;

      end;

   until (fin_camino=TRUE);

   // liberamos la lista de nodos alcanzables
   Finalize(l_nodos_alc);

   Result:= camino_completo;

end;


//------------------------------------------------------------------------------

(*
   Procedimiento que evala una solucin, es decir, calcula el fitness para cada
   uno de los objetivos. Los valores obtenidos los guarda en las variables de
   fitness de la estructura solucin.
   Recibe como parmetros:
      - sol -> la solucin (camino) a evaluar
      - recursos -> recursos disponibles para completar el camino
      - salud -> recursos disponibles para completar el camino
*)
procedure evaluar_solucion(var sol: t_solucion4; recursos: real; salud: real);

var
  i, num_nodos: integer;

  con_rec,con_sal: real;

  nodo_act, nodo_sig: t_celda;  // nodos actual y siguiente en el camino

begin

   // el nodo inicial es el actual
   nodo_act:= sol.camino[0];

   // nmero de nodos del camino
   num_nodos:= high(sol.camino)+1;


   // inicializaciones
   sol.fit_obj1:= 0;
   sol.fit_obj2:= 0;
   sol.fit_obj3:= 0;
   sol.fit_obj4:= 0;
   sol.fit_obj1_2o:= 0;
   sol.fit_obj2_2o:= 0;
   sol.consum_rec:= 0;
   sol.consum_sal:= 0;

   // recorremos los nodos del camino
   for i:= (low(sol.camino)+1) to high(sol.camino) do begin

       // nodo siguiente del camino
       nodo_sig:= sol.camino[i];

       // consumos de recursos y salud
       con_rec:= consumo_recursos(nodo_act,nodo_sig,recursos);
       con_sal:= consumo_salud(nodo_act,nodo_sig,salud);

       // Fitness Objetivo 1 (Consumo de Recursos)
       sol.fit_obj1:= sol.fit_obj1 + con_rec;
       sol.consum_rec:= sol.consum_rec + con_rec;

       // Fitness Objetivo 2 (Distancia Eucldea Media al Destino)
       sol.fit_obj2:= sol.fit_obj2 +
                     (distancia_eu(Tablero[nodo_sig.y,nodo_sig.x],NODO_FINAL)/(num_nodos-1));

       // Fitness Objetivo 3 (Consumo de Salud)
       sol.fit_obj3:= sol.fit_obj3 + con_sal;
       sol.consum_sal:= sol.consum_sal + con_sal;

       // Fitness Objetivo 4 (Visibilidad de las Celdas del camino)
       sol.fit_obj4:= sol.fit_obj4 + (1 - ZonasO[nodo_sig.y,nodo_sig.x]);


       // Calculamos los Fitness de las soluciones para 2 objetivos
       sol.fit_obj1_2o:= sol.fit_obj1_2o + con_rec
                                         + (KEV1_ZVO * (1 - ZonasO[nodo_sig.y,nodo_sig.x]));
       sol.fit_obj2_2o:= sol.fit_obj2_2o + con_sal
                                         + (KEV2_ZVO * (1 - ZonasO[nodo_sig.y,nodo_sig.x]));

       // el nuevo nodo actual es el siguiente de esta iteracin
       nodo_act:= nodo_sig;

   end;

end;

//------------------------------------------------------------------------------

(*
   Funcin que comprueba si una solucin es suficientemente buena como
   para incluirla en el Frente de Pareto, es decir, si no es dominada por ninguna otra.
   Recibe como parmetros:
      - sol -> la solucin (camino) a comprobar
   Devuelve como resultado:
      - TRUE -> si la solucin no es dominada por ninguna del Frente de Pareto
      - FALSE -> si la solucin es dominada por alguna
*)
function FP_comprobar_solucion(sol: t_solucion4):boolean;

var
  i: integer;
  dominada: boolean;

begin

   if (Frente_Pareto.num_sols_FP = 0) then begin
      Result:= TRUE;
      exit;
   end;

   dominada:= FALSE;

   // recorremos el Frente de Pareto Comprobando si la solucin es dominada por alguna
   // (una solucin domina a otra si es mejor en ambos objetivos o es igual en uno y
   //  mejor en otro)
   // en este caso es mejor el de menor valor
   for i:=0 to (Frente_Pareto.num_sols_FP-1) do begin

       if ((sol.fit_obj1 >  Frente_Pareto.soluciones_FP[i].fit_obj1) and
           (sol.fit_obj2 >= Frente_Pareto.soluciones_FP[i].fit_obj2) and
           (sol.fit_obj3 >= Frente_Pareto.soluciones_FP[i].fit_obj3) and
           (sol.fit_obj4 >= Frente_Pareto.soluciones_FP[i].fit_obj4)) or
          ((sol.fit_obj1 >= Frente_Pareto.soluciones_FP[i].fit_obj1) and
           (sol.fit_obj2 >  Frente_Pareto.soluciones_FP[i].fit_obj2) and
           (sol.fit_obj3 >= Frente_Pareto.soluciones_FP[i].fit_obj3) and
           (sol.fit_obj4 >= Frente_Pareto.soluciones_FP[i].fit_obj4)) or
          ((sol.fit_obj1 >= Frente_Pareto.soluciones_FP[i].fit_obj1) and
           (sol.fit_obj2 >= Frente_Pareto.soluciones_FP[i].fit_obj2) and
           (sol.fit_obj3 >  Frente_Pareto.soluciones_FP[i].fit_obj3) and
           (sol.fit_obj4 >= Frente_Pareto.soluciones_FP[i].fit_obj4)) or
          ((sol.fit_obj1 >= Frente_Pareto.soluciones_FP[i].fit_obj1) and
           (sol.fit_obj2 >= Frente_Pareto.soluciones_FP[i].fit_obj2) and
           (sol.fit_obj3 >= Frente_Pareto.soluciones_FP[i].fit_obj3) and
           (sol.fit_obj4 >  Frente_Pareto.soluciones_FP[i].fit_obj4)) then begin
           // es dominada por alguna otra solucin
           dominada:= TRUE;
           break;
       end
       else if (sol.fit_obj1 = Frente_Pareto.soluciones_FP[i].fit_obj1) and
               (sol.fit_obj2 = Frente_Pareto.soluciones_FP[i].fit_obj2) and
               (sol.fit_obj3 = Frente_Pareto.soluciones_FP[i].fit_obj3) and
               (sol.fit_obj4 = Frente_Pareto.soluciones_FP[i].fit_obj4) then begin
           // es igual que alguna otra solucin (la consideramos dominada para
           // no repetirla)
           dominada:= TRUE;
           break;
       end;

   end;

   Result:= (not dominada);

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que incluye una solucin en el Frente de Pareto, eliminando del
   mismo aquellas que sean dominadas por ella.
   Recibe como parmetros:
      - sol -> la solucin (camino) a incluir
*)
procedure FP_incluir_solucion(sol: t_solucion4);

var
  i: integer;
  FP_aux: array of t_solucion4;
  num_fp_aux: integer;

begin

   // nmero de elementos en el FP auxiliar
   num_fp_aux:= 1;
   // se copia la solucin a incluir en el FP auxiliar
   SetLength(FP_aux,num_fp_aux);
   FP_aux[0]:= sol;

   // recorremos el Frente de Pareto copiando las soluciones no dominadas al
   // FP auxiliar
   for i:=0 to (Frente_Pareto.num_sols_FP-1) do begin

       if ((sol.fit_obj1 <  Frente_Pareto.soluciones_FP[i].fit_obj1) and
           (sol.fit_obj2 <= Frente_Pareto.soluciones_FP[i].fit_obj2) and
           (sol.fit_obj3 <= Frente_Pareto.soluciones_FP[i].fit_obj3) and
           (sol.fit_obj4 <= Frente_Pareto.soluciones_FP[i].fit_obj4)) or
          ((sol.fit_obj1 <= Frente_Pareto.soluciones_FP[i].fit_obj1) and
           (sol.fit_obj2 <  Frente_Pareto.soluciones_FP[i].fit_obj2) and
           (sol.fit_obj3 <= Frente_Pareto.soluciones_FP[i].fit_obj3) and
           (sol.fit_obj4 <= Frente_Pareto.soluciones_FP[i].fit_obj4)) or
          ((sol.fit_obj1 <= Frente_Pareto.soluciones_FP[i].fit_obj1) and
           (sol.fit_obj2 <= Frente_Pareto.soluciones_FP[i].fit_obj2) and
           (sol.fit_obj3 <  Frente_Pareto.soluciones_FP[i].fit_obj3) and
           (sol.fit_obj4 <= Frente_Pareto.soluciones_FP[i].fit_obj4)) or
          ((sol.fit_obj1 <= Frente_Pareto.soluciones_FP[i].fit_obj1) and
           (sol.fit_obj2 <= Frente_Pareto.soluciones_FP[i].fit_obj2) and
           (sol.fit_obj3 <= Frente_Pareto.soluciones_FP[i].fit_obj3) and
           (sol.fit_obj4 <  Frente_Pareto.soluciones_FP[i].fit_obj4)) then begin
           // domina a alguna otra solucin (se pasa a la siguiente sin copiarla)
           continue;
       end;

       // se copia la solucin actual al FP auxiliar
       num_fp_aux:= num_fp_aux+1;
       SetLength(FP_aux,num_fp_aux);

       FP_aux[num_fp_aux-1]:= Frente_Pareto.soluciones_FP[i];

   end;

   // se copia el FP auxiliar al Frente de Pareto real
   Finalize(Frente_Pareto.soluciones_FP);

   SetLength(Frente_Pareto.soluciones_FP,num_fp_aux);

   for i:=0 to (num_fp_aux-1) do
      Frente_Pareto.soluciones_FP[i]:= FP_aux[i];

   Frente_Pareto.num_sols_FP:= num_fp_aux;

   // liberamos la memoria del FP auxiliar
   Finalize(FP_aux);

end;


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

(*
   Procedimiento que guarda los valores de la mejor solucin que se ha encontrado
   segn el criterio con el que se busc.
   Recibe como parmetros:
      - fich_mapa -> nombre del fichero del mapa
      - identif -> cadena que identifica al fichero
      - semilla -> semilla usada para los aleatorios
      - valf1obj2 -> valor del fitness del objetivo 1, como cadena
      - valf2obj2 -> valor del fitness del objetivo 2, como cadena
      - valf1 -> valor del fitness del objetivo 1, como cadena
      - valf2 -> valor del fitness del objetivo 2, como cadena
      - valf3 -> valor del fitness del objetivo 3, como cadena
      - valf4 -> valor del fitness del objetivo 4, como cadena
*)
procedure guardar_minimos(fich_mapa: string; identif: string; semilla: string;
                          valf1obj2: string; valf2obj2:string;
                          valf1: string; valf2:string; valf3: string; valf4:string);

var
   nombre_fich: string;
   fich_mins: TextFile;

begin

      // se crea el nombre
      nombre_fich:= '[' + ChangeFileExt(ExtractFileName(fich_mapa),'') + '] ';
      nombre_fich:= nombre_fich + 'BestF' + identif + '_' + SEL_SIG_NODO + '_' +
                    APLIC_LAMBDA + '_' + IntToStr(N_ITERACIONES) + 'it' + IntToStr(N_HORMIGAS) + 'a';

      if (HACER_BL=TRUE) then
         nombre_fich:= nombre_fich + '_ls';

      if (LF=0.5) then
         nombre_fich:= nombre_fich + '_rs'
      else if (LF<0.5) then
         nombre_fich:= nombre_fich + '_+s'
      else
         nombre_fich:= nombre_fich + '_+r';

      nombre_fich:= nombre_fich + '.txt';

      // se comprueba si el fichero existe
      if not (FileExists(nombre_fich)) then begin
         // si no existe, se crea y se mete una lnea de informacin de los datos
         AssignFile(fich_mins, nombre_fich);
         Rewrite(fich_mins);

         // Valores de los Parmetros
         Writeln(fich_mins, 'Parameters');
         Writeln(fich_mins, '----------');
         Writeln(fich_mins, 'ALPHA=' + FloatToStr(ALPHA) +
                            '  BETA=' + FloatToStr(BETA) +
                            '  LAMBDA=' + FloatToStr(LF) +
                            '(' + APLIC_LAMBDA + ')' );
         Writeln(fich_mins, 'RHO=' + FloatToStr(RHO) +
                            '  q0=' + FloatToStr(q0));
         Writeln(fich_mins, '------------------------------------');
         Writeln(fich_mins, 'UNIT:');
         Writeln(fich_mins, 'resources = ' + FloatToStr(RECURSOS_H) +
                            '  health = ' + FloatToStr(SALUD_H) +
                            '  Adquisition Capability = ' + IntToStr(CAP_ADQ_UNI));
         Writeln(fich_mins, '------------------------------------');
         Writeln(fich_mins, 'ENEMY:');
         Writeln(fich_mins, 'Adquisition Capability = ' + IntToStr(CAP_ADQ_ENE) +
                            '  num. known enemies = ' + IntToStr(num_enemigos_conocidos));
         Writeln(fich_mins, '---');
         Writeln(fich_mins, '---');

         // Descripcin de columnas
         Writeln(fich_mins, 'Seed' + #9 + 'Ff Cost' + #9 + 'Fs Cost');
      end
      else begin
         // si existe, se abre para aadir datos
         AssignFile(fich_mins, nombre_fich);
         Append(fich_mins);
      end;

      // se guardan los valores mnimos con la semilla (separados por tabulador)
      Writeln(fich_mins, semilla + #9 + valf1obj2 + #9 + valf2obj2 + #9 + '|4objs|' +
                                   #9 + valf1 + #9 + valf2 + #9 + valf3 + #9 + valf4);
      Flush(fich_mins);
      CloseFile(fich_mins);

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que realiza una bsqueda local sobre una solucin
   Recibe como parmetros:
      - sol -> la solucin (camino) a refinar
      - recursos -> recursos disponibles para completar el camino
      - salud -> recursos disponibles para completar el camino
*)
procedure busqueda_local(var sol: t_solucion4; recursos: real; salud: real);

var
  ind,ind_c,i,j,k,l: integer;
  num_nodos, num_alc, num_comunes: integer;

  l_nodos_alc_act,l_nodos_alc_sig,l_nodos_alc_ant: t_alcanzables;  // lista de vecinos alcanzables

  l_comunes: array of t_celda;  // lista de nodos comunes a los dos vecindarios (interseccin)
  nodo_bak: t_celda;
  fit1_bak, fit2_bak, fit3_bak, fit4_bak: real;
  conrec_bak, consal_bak: real;

  rec_cons, sal_cons: real;

begin

    // nmero de nodos del camino
    num_nodos:= high(sol.camino)+1;

    for i:=1 to N_ITERACS_BL do begin

        // elegimos aleatoriamente el nodo a cambiar (dentro del camino)
        // no se elige ni el primero ni el ltimo (origen y fin son fijos)
        ind:=Random((num_nodos-2)) + 1;

        (* se crea la lista de vecinos alcanzables para el nodo elegido,
         para el siguiente y para el anterior del camino *)
        Finalize(l_nodos_alc_act);
        // calculamos los recursos y salud consumida hasta el nodo en cuestin (para ver los alcanzables)
        rec_cons:= consumo_recursos_camino(sol.camino,0,ind);
        sal_cons:= consumo_salud_camino(sol.camino,0,ind);
        num_alc:= lista_alcanzables(sol.camino[ind],(salud-sal_cons),(recursos-rec_cons),TRUE,sol.camino,num_nodos,l_nodos_alc_act);
        if (num_alc=0) then continue;

        Finalize(l_nodos_alc_sig);
        // calculamos los recursos y salud consumida hasta el nodo en cuestin (para ver los alcanzables) [se mira hasta el nodo anterior]
        rec_cons:= consumo_recursos_camino(sol.camino,0,ind);
        sal_cons:= consumo_salud_camino(sol.camino,0,ind);
        num_alc:= lista_alcanzables(sol.camino[ind+1],(salud-sal_cons),(recursos-rec_cons),TRUE,sol.camino,num_nodos,l_nodos_alc_sig);
        if (num_alc=0) then continue;

        Finalize(l_nodos_alc_ant);
        // calculamos los recursos y salud consumida hasta el nodo en cuestin (para ver los alcanzables) [se mira hasta el nodo siguiente]
        rec_cons:= consumo_recursos_camino(sol.camino,0,ind);
        sal_cons:= consumo_salud_camino(sol.camino,0,ind);
        num_alc:= lista_alcanzables(sol.camino[ind-1],(salud-sal_cons),(recursos-rec_cons),TRUE,sol.camino,num_nodos,l_nodos_alc_ant);
        if (num_alc=0) then continue;

        (* creamos lista de nodos comunes a los 3 vecindarios *)
        num_comunes:= 0;
        Finalize(l_comunes);

        for j:= low(l_nodos_alc_act) to high(l_nodos_alc_act) do begin
           for k:= low(l_nodos_alc_sig) to high(l_nodos_alc_sig) do begin
              for l:= low(l_nodos_alc_ant) to high(l_nodos_alc_ant) do begin

                 // si un nodo est en los tres vecindarios, se guarda en la lista de comunes
                 if ((l_nodos_alc_act[j].x = l_nodos_alc_sig[k].x) and (l_nodos_alc_act[j].x = l_nodos_alc_ant[l].x))
                    and
                    ((l_nodos_alc_act[j].y = l_nodos_alc_sig[k].y) and (l_nodos_alc_act[j].y = l_nodos_alc_ant[l].y)) then begin

                    SetLength(l_comunes, num_comunes+1);
                    l_comunes[num_comunes]:= l_nodos_alc_act[j];
                    num_comunes:= num_comunes+1;

                 end;

              end;
           end;
        end;

        // si hay nodos comunes, se elige uno de ellos y sustituye al nodo actual
        if (num_comunes>0) then begin

           // se hace una copia de la celda a cambiar en la solucin y el
           // fitness de esta
           nodo_bak:= sol.camino[ind];
           fit1_bak:= sol.fit_obj1;
           fit2_bak:= sol.fit_obj2;
           fit3_bak:= sol.fit_obj3;
           fit4_bak:= sol.fit_obj4;

           conrec_bak:= sol.consum_rec;
           consal_bak:= sol.consum_sal;

           // se cambia la celda por otra
           ind_c:=Random(num_comunes);
           sol.camino[ind]:= l_comunes[ind_c];

           // se evala la nueva solucin (con la celda distinta)
           evaluar_solucion(sol,recursos,salud);

           // si es dominada por la anterior, se mantiene la solucin sin cambiar
           if ((sol.fit_obj1 > fit1_bak) and
               (sol.fit_obj2 >= fit2_bak) and
               (sol.fit_obj3 >= fit3_bak) and
               (sol.fit_obj4 >= fit4_bak)) or
              ((sol.fit_obj1 >= fit1_bak) and
               (sol.fit_obj2 >  fit2_bak) and
               (sol.fit_obj3 >= fit3_bak) and
               (sol.fit_obj4 >= fit4_bak)) or
              ((sol.fit_obj1 >= fit1_bak) and
               (sol.fit_obj2 >= fit2_bak) and
               (sol.fit_obj3 >  fit3_bak) and
               (sol.fit_obj4 >= fit4_bak)) or
              ((sol.fit_obj1 >= fit1_bak) and
               (sol.fit_obj2 >= fit2_bak) and
               (sol.fit_obj3 >= fit3_bak) and
               (sol.fit_obj4 >  fit4_bak)) or
              ((sol.fit_obj1 = fit1_bak) and
               (sol.fit_obj2 = fit2_bak) and
               (sol.fit_obj3 = fit3_bak) and
               (sol.fit_obj4 = fit4_bak)) then begin

              sol.camino[ind]:= nodo_bak;
              sol.fit_obj1:= fit1_bak;
              sol.fit_obj2:= fit2_bak;
              sol.consum_rec:= conrec_bak;
              sol.consum_sal:= consal_bak;

           end;

        end;

    end;  // bsqueda local

end;


//------------------------------------------------------------------------------

(* Algoritmo Principal CHAC *)

procedure Alg_CHAC_4objs ();

var
  //contadores
  n,h: integer;

  sol_valida: boolean;
  solucs_validas: integer;

  // T0 indican la Bondad de las soluciones actuales y de la iteracin
  // T0_it: real;

begin

   // Inicializaciones
   (*
    Solucin inicial factible (Greedy por ejemplo)
    calcular T0 de las matrices de Feromonas a partir de los fitness de la solucin

    T0 := 1.0/(num_nodos * Fitnes_Obj1 * Fitnes_Obj2)
    T1[i][j] := 1.0/(num_nodos * Fitnes_Obj1)
    T2[i][j] := 1.0/(num_nodos * Fitnes_Obj2)
   *)

   // Valores iniciales de feromona
   // Mximos recursos que consume una celda. Feromona inicial del Objetivo 1 (Recursos)
   T01:= 1.0/((N_FIL_TAB*N_COL_TAB)*MAX_RECURSOS);
   // Mxima distancia desde una celda al final (diagonal). Feromona inicial del Objetivo 2 (Distancia)
   T02:= 1.0/((N_FIL_TAB*N_COL_TAB)*SQRT(SQR(N_FIL_TAB)+SQR(N_COL_TAB)));
//   T02:= T02/10.0;
   // Mxima salud que consume una celda. Feromona inicial del Objetivo 3 (Salud)
   T03:= 1.0/((N_FIL_TAB*N_COL_TAB)*MAX_DANIO);
   // Visibilidad de todas las celdas. Feromona inicial del Objetivo 4 (Visibilidad)
   T04:= 1.0/((N_FIL_TAB*N_COL_TAB)*1);

   // se inicializan las matrices con un valor para todos los arcos posibles
   inicializar_feromonas(T01,T02,T03,T04);

   // --------- BUCLE PRINCIPAL ---------
   n:=0;
   While (n<N_ITERACIONES) do begin

       dIT:= n;    //depuracin

       // Inicializaciones
       solucs_validas:= 0;

       inicializar_hormigas(SALUD_H,RECURSOS_H);

       // -- Creamos la solucin con cada hormiga --

       for h:=0 to (N_HORMIGAS-1) do begin

          dH:= h;   //depuracin

          // Construir Solucin para la hormiga h
          sol_valida:= construir_solucion(Hormigas[h],NODO_INICIAL,NODO_FINAL);

          // slo se hacen los pasos siguientes si la solucin es vlida
          if sol_valida=TRUE then begin

             solucs_validas:= solucs_validas + 1;

             // evaluamos la solucin obtenida
             evaluar_solucion(Hormigas[h].solucion,RECURSOS_H,SALUD_H);

             (* si se desea, se hace Bsqueda Local *)
             if (HACER_BL=TRUE) then begin

                busqueda_local(Hormigas[h].solucion,RECURSOS_H,SALUD_H);

             end;

             // Comprobar si la solucin pertenece al Frente de Pareto (si no es
             // dominada por ninguna del FP)
             if (FP_comprobar_solucion(Hormigas[h].solucion)=TRUE) then begin
                // si domina, se incluye en el frente y se eliminan las dominadas
                FP_incluir_solucion(Hormigas[h].solucion);
             end;

          end;

       end;  // for hormigas

       // Actualizacin Global de Feromona
       actualizacion_global_feromona();

       // pasamos a la siguiente iteracin
       n:= n+1;

   End;  // while principal

(*
   // aplicamos lambda en este momento (si se eligi hacerlo al final)
   if (APLIC_LAMBDA = 'pLBD') then begin
      showmessage('filtro final');
   end;
*)

end;

// --------------------------------------------------------------

(*
   Procedimiento que escribe el fichero con los datos de las soluciones
   Recibe como parmetros:
      - nombre_fichero -> fichero en el que se guardarn los datos
*)
procedure guardar_soluciones_2obj(nombre_fichero: string);
var
   fich_sols: File of t_celda;
   celda_aux: t_celda;
   i,j: integer;

begin

  // creamos el nombre del fichero como el nombre del mapa con extensin 'sol'
  //nombre_fichero:= ChangeFileExt(nombre_fichero,'.sol');

  // se abre el fichero para escritura
  AssignFile(fich_sols, nombre_fichero);
  ReWrite(fich_sols);

  // guardamos una celda con:
  // el nmero de soluciones (en x)
  celda_aux.x:= Frente_Pareto.num_sols_FP;
  // el valor de LF (en danio_arma)
  celda_aux.danio_arma:= LF;

  Write(fich_sols,celda_aux);


  // guardamos las soluciones encontradas
  for i:= low(Frente_Pareto.soluciones_FP) to high(Frente_Pareto.soluciones_FP) do begin

     // guardamos los datos de la solucin (nmero de nodos, fitness salud, fitness recursos)
     celda_aux.x:= high(Frente_Pareto.soluciones_FP[i].camino)+1;
     celda_aux.recursos_0:= Frente_Pareto.soluciones_FP[i].fit_obj1_2o;
     celda_aux.danio_0:= Frente_Pareto.soluciones_FP[i].fit_obj2_2o;
     Write(fich_sols,celda_aux);

     for j:= low(Frente_Pareto.soluciones_FP[i].camino) to
             high(Frente_Pareto.soluciones_FP[i].camino) do begin

        // guardamos los nodos del camino solucin
        Write(fich_sols,Frente_Pareto.soluciones_FP[i].camino[j]);

     end;

  end;

  // cerramos el fichero
  CloseFile(fich_sols);

  if (VARIAS_EJECUCIONES=FALSE) then
     showmessage(IntTostr(Frente_Pareto.num_sols_FP) + ' Solutions Found and Saved. Close the Parameters Window to see them');

end;


// --------------------------------------------------------------

(*
   Procedimiento que refina las soluciones encontradas eliminando celdas innecesarias
   (celdas intermedias entre dos celdas comunicadas entre si)
*)
procedure refinar_soluciones();
var
   celda_act, celda_sig: t_celda;
   i,j,k: integer;

   num_cels, num_sols, num_sols_ref: integer;

   sol_refinada: t_solucion4;
   cont_cels_ref: integer;

   FP_aux: array of t_solucion4;

begin

  // nmero de soluciones inicial
  num_sols:= high(Frente_Pareto.soluciones_FP)+1;
  num_sols_ref:= 0;

  // recorremos las soluciones encontradas
  for i:= low(Frente_Pareto.soluciones_FP) to (num_sols-1) do begin

     // inicializaciones
     num_cels:= high(Frente_Pareto.soluciones_FP[i].camino)+1;
     j:= low(Frente_Pareto.soluciones_FP[i].camino);

     // si el camino solo tiene 2 celdas no hay nada que refinar
     if (num_cels <= 2) then
       continue;

     // se inicializa la memoria de la solucin refinada
     cont_cels_ref:= 0;
     sol_refinada.fit_obj1:= 0;
     sol_refinada.fit_obj2:= 0;
     sol_refinada.fit_obj3:= 0;
     sol_refinada.fit_obj4:= 0;
     sol_refinada.fit_obj1_2o:= 0;
     sol_refinada.fit_obj2_2o:= 0;
     Finalize(sol_refinada.camino);

     // leemos la celda actual
     celda_act:= Frente_Pareto.soluciones_FP[i].camino[j];

     // se recorren las celdas del camino solucin
     while (j < num_cels) do begin

        // se aade la celda actual al camino refinado (seguro que est bien)
        SetLength(sol_refinada.camino,cont_cels_ref+1);
        sol_refinada.camino[cont_cels_ref]:= celda_act;
        cont_cels_ref:= cont_cels_ref + 1;

        // se miran las celdas empezando por el final y se busca el siguiente vecino
        // de la celda actual para ver si hay alguno ms adelante de la siguiente (a 1 paso)
        k:= num_cels-1;
        celda_sig:= Frente_Pareto.soluciones_FP[i].camino[k];
        while ((k >= j+1) and
               (es_vecino_hexa(celda_act.y,celda_act.x,celda_sig.y,celda_sig.x,N_FIL_TAB,N_COL_TAB)=FALSE))
               do begin
           k:= k-1;
           celda_sig:= Frente_Pareto.soluciones_FP[i].camino[k];
        end;

        // se pone el ltimo vecino encontrado como celda actual
        j:= k;
        celda_act:= Frente_Pareto.soluciones_FP[i].camino[j];

        // si es la ltima, se guarda
        if (j+1 >= num_cels) then begin
          SetLength(sol_refinada.camino,cont_cels_ref+1);
          sol_refinada.camino[cont_cels_ref]:= celda_act;
          cont_cels_ref:= cont_cels_ref + 1;
        end;

        j:= j+1;

     end;  // while j

     // si la nueva solucin tiene menos celdas se evala y se guarda en el FP_auxiliar
     if (cont_cels_ref < num_cels) then begin

        evaluar_solucion(sol_refinada,RECURSOS_H,SALUD_H);

        // se copia la solucin a incluir en el FP auxiliar
        SetLength(FP_aux,num_sols_ref+1);
        FP_aux[num_sols_ref]:= sol_refinada;

        num_sols_ref:= num_sols_ref+1;

     end;

  end;  // for i

  // Se guardan las soluciones refinadas en el FP (considerando las dominadas)
  if (num_sols_ref > 0) then begin

      i:= high(FP_aux);
      while (i >= low(FP_aux)) do begin
         // Comprobar si la solucin pertenece al Frente de Pareto (si no es
         // dominada por ninguna del FP)
         if (FP_comprobar_solucion(FP_aux[i])=TRUE) then begin
            // si domina, se incluye en el frente y se eliminan las dominadas
            FP_incluir_solucion(FP_aux[i]);
         end;

         i:= i-1;
      end;

      (*ShowMessage(IntTostr(num_sols_ref) + ' Soluciones Refinadas');*)

  end;

end;

// --------------------------------------------------------------

(*
   Procedimiento que ordena las soluciones encontradas en base al criterio elegido.
   Recibe como parmetros:
      - rs -> cadena que indica el criterio de ordenacin. '+r' segn recursos y
              '+s' segn salud (o compromiso r/s)
*)
procedure ordenar_soluciones_2obj(rs: string);
var
   i,j: integer;
   num_sols: integer;
   sol_aux: t_solucion4;

   min_r, min_s: real;
   ind_min: integer;

begin

  // nmero de soluciones inicial
  num_sols:= high(Frente_Pareto.soluciones_FP)+1;

  // Ordenamos por el mtodo de Seleccin
  for i:= low(Frente_Pareto.soluciones_FP) to (num_sols-1) do begin

     min_r:= Frente_Pareto.soluciones_FP[i].fit_obj1_2o;
     min_s:= Frente_Pareto.soluciones_FP[i].fit_obj2_2o;
     ind_min:= i;

     for j:= i+1 to (num_sols-1) do begin

        // mnimos recursos
        if rs='+r' then begin

           if (min_r > Frente_Pareto.soluciones_FP[j].fit_obj1_2o) then begin
              // guardamos los fitness como mnimos
              min_r:= Frente_Pareto.soluciones_FP[j].fit_obj1_2o;
              min_s:= Frente_Pareto.soluciones_FP[j].fit_obj2_2o;
              ind_min:= j;
           end;

        end
        // mnima salud (o compromiso r/s)
        else begin

           if (min_s > Frente_Pareto.soluciones_FP[j].fit_obj2_2o) then begin
              // guardamos los fitness como mnimos
              min_r:= Frente_Pareto.soluciones_FP[j].fit_obj1_2o;
              min_s:= Frente_Pareto.soluciones_FP[j].fit_obj2_2o;
              ind_min:= j;
           end;

        end;

     end;   // for j

     // se intercambia el menor con el i
     if (i<>ind_min) then begin
        // se reserva memoria
        // SetLength(sol_aux.camino,High(Frente_Pareto.soluciones_FP[i].camino));
        sol_aux:= Frente_Pareto.soluciones_FP[i];
        Frente_Pareto.soluciones_FP[i]:= Frente_Pareto.soluciones_FP[ind_min];
        Frente_Pareto.soluciones_FP[ind_min]:= sol_aux;
     end;

  end;   // for i


end;

// -----------------------------------------------------------------------------


(*              FUNCIN PRINCIPAL de CHAC                *)
Function Principal_CHAC_4objs (fich_mapa: string; vlf: real; aplic_lbd: string; DisZVO: integer; sel_nodo: string;
                               salud_cia: real; recursos_cia: real; num_it: integer;
                               num_horm: integer; busq_local: boolean; iteracs_bl: integer;
                               refinar: boolean; SEM: integer; enem_conocido: boolean; pasos_busq_enems: integer;
                               fich_sols: string; varias_ejecs: boolean; dat_estadisticas: boolean): integer;

begin

   // Inicializamos la semilla para los nmeros aleatorios
   RandSeed:= SEM;

   // inicializamos las 'constantes' de programa
   RECURSOS_H:= recursos_cia;
   SALUD_H:= salud_cia;
   N_ITERACIONES:= num_it;
   N_HORMIGAS:= num_horm;
   DVO:= DisZVO;
   LF:= vlf;
   APLIC_LAMBDA:= aplic_lbd;

   HACER_BL:= busq_local;
   N_ITERACS_BL:= iteracs_bl;

   SEL_SIG_NODO:= sel_nodo;

   ENEMIGOS_CONOCIDOS:= enem_conocido;
   PASOS_BUSCAR_ENEMIGOS:=  pasos_busq_enems;

   VARIAS_EJECUCIONES:= varias_ejecs;

   // inicializamos los nodos con valores extremos por control
   NODO_INICIAL.altura:= High(Integer);
   NODO_FINAL.altura:= High(Integer);

   // Si no hay enemigos, se considerar mucho ms el haber ido por celdas con mayor
   // radio de ocultacin
   //if (num_enemigos_conocidos=0) then begin
   //   KN2_VO:= 0.7;
   //   KN2_DIS:= 0.2;
   //end;

   // leemos el fichero de datos y creamos las matrices del programa
   cargar_inicializar_datos(fich_mapa);

   // comprobamos si se han leido bien los datos
   if ((NODO_INICIAL.altura = High(Integer)) or
       (NODO_FINAL.altura = High(Integer))) then begin

       // no se han leido correctamente los puntos de Origen y Destino
       ShowMessage('There is no Origin or Destination Point in the Map saved in the File');

       Result:= 0;

       // liberamos la memoria
       liberar_memoria();

       exit;

   end;

   // lanzamos el Algoritmo
   Alg_CHAC_4objs();

   // apuntamos el nmero de soluciones encontradas
   Result:= Frente_Pareto.num_sols_FP;

   if (Result>0) then begin

      // si se desea, se refinan las soluciones eliminando bucles innecesarios
      // (celdas intermedias no necesarias)
      if (refinar=TRUE) then
         refinar_soluciones();

      // apuntamos el nmero de soluciones encontradas + refinadas
      Result:= Frente_Pareto.num_sols_FP;

      // Si se hace con RD, slo habr un FP, no hay uno +r y otro +s)
      if (SEL_SIG_NODO = 'RT') then begin

        // ordenamos las soluciones en base al criterio al que mayor prioridad
        // se haya dado
        if (LF<=0.5) then
           ordenar_soluciones_2obj('+s')    // se ordenan en base a la seguridad
        else
           ordenar_soluciones_2obj('+r');   // se ordenan en base a la rapidez

        // copiamos las soluciones a un fichero
        guardar_soluciones_2obj(fich_sols);

        // si se desea, se guardan los mejores costes en un fichero aparte
        guardar_minimos(fich_mapa,'_CHAC4',IntToStr(SEM),
                        FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj1_2o),FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj2_2o),
                        FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj1),FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj2),
                        FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj3),FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj4));
      end
      else begin
            LF := 0.5;  // simplemente para dar un nombre coherente al fichero de mnimos

            ordenar_soluciones_2obj('+r');   // se ordenan en base a la rapidez

            // si se desea, se guardan los mejores costes en un fichero aparte
            guardar_minimos(fich_mapa,'_CHAC4_(+r)',IntToStr(SEM),
                            FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj1_2o),FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj2_2o),
                            FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj1),FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj2),
                            FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj3),FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj4));

            ordenar_soluciones_2obj('+s');   // se ordenan en base a la seguridad

            // si se desea, se guardan los mejores costes en un fichero aparte
            guardar_minimos(fich_mapa,'_CHAC4_(+s)',IntToStr(SEM),
                            FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj1_2o),FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj2_2o),
                            FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj1),FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj2),
                            FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj3),FloatToStr(Frente_Pareto.soluciones_FP[0].fit_obj4));

            // copiamos las soluciones a un fichero
            guardar_soluciones_2obj(fich_sols);

      end;

   end
   else begin
       if (VARIAS_EJECUCIONES=FALSE) then
          ShowMessage('No solution has been found');

       // se guarda una raya para indicar 'no solucin'
       guardar_minimos(fich_mapa,'_CHAC',IntToStr(SEM),'-','-','-','-','-','-');
   end;

   // liberamos la memoria
   liberar_memoria();

end;



end.
