unit Simulador;

interface

(* ****************** CONSTANTES DE SIMULACIN ********************** *)

const

  (*
  // capacidad de adquisicin de la unidad en nmero de celdas (0 = no ve nada)
  CAP_ADQ_UNI: integer = 10;   // 9 Kilmetros <=> 18 celdas

  // capacidad de adquisicin del enemigo en nmero de celdas (0 = no ve nada)
  CAP_ADQ_ENE: integer = 10;  // 9 Kilmetros <=> 18 celdas

  // mxima diferencia de alturas que puede atravesar la unidad
  MAX_DIF_ALTURAS_UNI: integer = 2;
  *)

  // dao que hacen las celdas por defecto (consumo de salud) [en unidades de salud]
  DANIO_0_NORM: single = 0.1;
  DANIO_0_BOSQ: single = 0.3;
  DANIO_0_AGUA: single = 0.5;
  DANIO_0_OBST: single = 0;

  // recursos que consumen las celdas por defecto [en unidades de recursos]
  RECURSOS_0_NORM: single = 1;
  RECURSOS_0_BOSQ: single = 3;
  RECURSOS_0_AGUA: single = 5;
  RECURSOS_0_OBST: single = 0;

  // intervalos de daos por letalidad en la celdas
  DANIO_NIVEL_1: single = 25;
  DANIO_NIVEL_2: single = 50;
  DANIO_NIVEL_3: single = 75;
  DANIO_NIVEL_4: single = 100;

  // Mximos
  { [*AJUSTE PARAMS*]
    MAX_DANIO: single = 100 }
  MAX_DANIO: single = 8;    // dao mximo que puede hacer una celda (80*menor valor de dao)
  MAX_RECURSOS: single = 5;   // recursos mximos consumidos al pasar por una celda


(* *************** TIPOS DEFINIDOS PARA EL SIMULADOR **************** *)

type

  // tipos de celdas
  t_tipo_celda = (AGUA,NORM,BOSQ,OBST);  // Agua, Normal, Bosque, Obstculo
  t_tipo_esp_celda = (NO,PORIGEN,PDESTINO,ENEMIGO,IMPARMA);  // Tipos especiales de Celdas

  // tipo que define una celda (las que se guardan en fichero de mapa)
  t_celda = Record
     x,y: integer;       // y <-> fila , x <-> columna
     altura: integer;
     tipo: t_tipo_celda;
     tipo_esp: t_tipo_esp_celda;
     danio_arma: real;   // dao por impacto de arma
     danio_0: real;      // dao por defecto de la celda (por tipo)
     recursos_0: real;   // recursos que consume el paso por ella (por tipo)
   end;

  // tipo que define un conjunto de celdas
  t_tablero = array of array of t_celda;

  // tipo que define una serie de alturas mximas permitidas en la lnea de vision
  t_alt_max_VO = array of integer;

  // tipo camino (lista de celdas)
  t_camino = array of t_celda;

  // tipo de array de enemigos
  t_lista_enemigos = array of t_celda;

  // ------------------- EVENTOS (DINAMISMO) ----------------------

  // tipos de eventos
  t_tipo_evento = (INS_ENE,   // Insertar enemigo
                   DEL_ENE,   // Borrar enemigo
                   MOV_ENE,   // Mover enemigo
                   MOV_DES);  // Mover destino

  // tipo que define un evento que se produce sobre el mapa
  t_evento = Record
     it: integer;              // iteracin en la que se produce el evento
     tipo: t_tipo_evento;      // tipo de evento
     x_act, y_act: integer;    // coordenadas actuales
     x_sig, y_sig: integer;    // coordenadas siguientes
   end;

  // lista de eventos
  t_lista_eventos = array of t_evento;

(* ********** VARIABLES GLOBALES (constantes) PBLICAS DEL SIMULADOR ********** *)

var

  // capacidad de adquisicin de la unidad en nmero de celdas (0 = no ve nada)
  CAP_ADQ_UNI: integer;   // 9 Kilmetros <=> 18 celdas

  // capacidad de adquisicin del enemigo en nmero de celdas (0 = no ve nada)
  CAP_ADQ_ENE: integer;  // 9 Kilmetros <=> 18 celdas

  // mxima diferencia de alturas que puede atravesar la unidad
  MAX_DIF_ALTURAS_UNI: integer;



(* ********** FUNCIONES Y PROCEDIMIENTOS PBLICOS DEL SIMULADOR ********** *)

function consumo_salud (celdai: t_celda; celdaf: t_celda; salud: real):real;
function consumo_recursos (celdai: t_celda; celdaf: t_celda; recursos: real):real;
function consumo_salud_camino (ruta: t_camino; ini: integer; fin:integer):real;
function consumo_recursos_camino (ruta: t_camino; ini: integer; fin:integer):real;
function celda_VO (celdai: t_celda; celdaf: t_celda; var zona: t_tablero; max_x: integer; max_y: integer):byte;
function hay_obstaculo(celdai: t_celda; celdaf: t_celda): boolean;
function celda_en_area_ca_hex(celdac: t_celda; celdau: t_celda; radio_ca: integer;
                              var Tablero: t_tablero; max_x: integer; max_y: integer): boolean;
function buscar_enemigos_ca_hex(celdau: t_celda; radio_ca: integer; var Enems: t_lista_enemigos;
                                var Tablero: t_tablero; max_x: integer; max_y: integer): integer;
function buscar_enemigos_ca_dist(celdau: t_celda; radio_ca: integer; var Enems: t_lista_enemigos;
                                 var Tablero: t_tablero; max_x: integer; max_y: integer; cad_dist: string): integer;
function dist_DDA (celdai: t_celda; celdaf: t_celda; var incx: real; var incy: real):integer;


implementation

(******************************************************************************)
(************************* FUNCIONES Y PROCEDIMIENTOS *************************)
(******************************************************************************)

//------------------------------------------------------------------------------

(*
   Procedimiento que calcula los incrementos en x e y que hay que hacer para
   seguir una recta segn el algoritmo DDA (Digital Differential Analizer)
   Recibe como parmetros:
      - xi -> x del punto inicial
      - yi -> y del punto inicial
      - xf -> x del punto final
      - yf -> y del punto final
   Devuelve:
      - IncX -> incremento en x que hay que hacer
      - IncY -> incremento en y que hay que hacer
*)
procedure linea_DDA (xi:integer; yi:integer; xf:integer; yf:integer;
                     var IncX: real; var IncY: real);
var
  dx, dy, paso: integer;

begin

   // diferencias
   dx:= xf-xi;
   dy:= yf-yi;

   // paso
   if abs(dx) > abs(dy) then paso:= abs(dx)
   else paso:= abs(dy);

   // incrementos
   IncX:= dx/paso;
   IncY:= dy/paso;

end;

//------------------------------------------------------------------------------

(*
   Funcin que calcula la distancia entre 2 celdas siguiendo una lnea DDA
   Recibe como parmetros:
      - celdai -> celda inicial
      - celdaf -> celda final
      - incx -> incremento en x para construir la recta DDA
      - incy -> incremento en y para construir la recta DDA
   Devuelve:
      - nmero de celdas entre las 2 dadas siguiendo una lnea DDA
*)
function dist_DDA (celdai: t_celda; celdaf: t_celda; var incx: real; var incy: real):integer;

var
  x,y: real;
  dist: integer;

begin

   if (incx=0) and (incy=0) then
      // calculamos los incrementos de x e y para la recta DDA
      linea_DDA(celdai.x,celdai.y,celdaf.x,celdaf.y,incx,incy);

   dist:= 0;

   // recorremos la lnea contando las celdas recorridas
   x:= celdai.x;
   y:= celdai.y;

   while ((round(x)<>celdaf.x) or (round(y)<>celdaf.y)) do begin
      dist:= dist + 1;
      // pasamos a la siguiente celda
      x:= x + incx;
      y:= y + incy;
   end;

   Result:= dist;

end;

//------------------------------------------------------------------------------

(*
   Procedimiento que obtiene un vector con las alturas mximas permitidas en las
   celdas intermedias de la lnea de visin para que la celdai sea vista desde
   celdaf (celda i ser la de menor altura)
   Recibe como parmetros:
      - celdai -> celda inicial de la comprobacin
      - celdaf -> celda final de la comprobacin
      - dist_if -> distancia entre la celda inicial y final a comprobar
   Devuelve:
      - amVO -> vector con las alturas mximas permitidas para cada distancia.
                El ndice indica la distancia a la celda i
*)
procedure calculo_alt_max_VO (celdai: t_celda; celdaf: t_celda; dist_if: integer;
                              var amVO: t_alt_max_VO);

var
    i, j, ind, cont: integer;
    med,max: integer;
    med_r: real;

    primer_ind, pos_med, signo: integer;

    hay_increm: boolean;

    repets: array of integer;

begin

    // inicializamos los valores medio y mximo
    med_r:= (1+(celdai.altura-celdaf.altura))/2;
    med:= round(int(med_r + 0.5));
    max:= celdai.altura-celdaf.altura;

    // construimos el vector de repeticiones (fundamental para crear el vector de alturas)
    // el ndice i corresponde a la altura a repetir en el vector de alturas mximas
    // el valor de repets[i] corresponde con el nmero de repeticiones en el vector de alturas
    // los ndices tienen correspondencia (max=0)...(min=N elementos), es decir repets[0]
    // sern las repeticiones del valor de altura mximo en el vector de alturas
    SetLength(repets,max);

    // si es par el primer ndice ser a la derecha del medio
    if (max mod 2) = 0 then
       primer_ind:= 1
    // si es impar el primer ndice ser a la izquierda del medio
    else
       primer_ind:= -1;

    pos_med:= (max-med);
    cont:= 1;  signo:= 1;
    while cont <= (dist_if-1) do begin

        // inicializamos el ndice la primera vez y cada 'max' pasos
        if ((cont mod max) = 1) or (max = 1) then begin
           ind:= 0;
           signo:= 1;
           hay_increm:= true;
        end
        else begin
            // pasamos al primer ndice
            if ((cont mod max) = 2) or (max = 2) then begin
                ind:= primer_ind;
                signo:= primer_ind;
            end;
        end;

        // incrementamos la repeticin del valor correspondiente
        repets[pos_med+ind]:= repets[pos_med+ind] + 1;

        if (hay_increm = true) then begin
            ind:= ind + (1 * signo);
            hay_increm:= false;
        end
        else begin
            hay_increm:= true;
        end;

        ind:= -ind; // le cambiamos el signo para acceder a la otra parte del vector
        signo:= -signo;
        cont:= cont + 1;

    end;


    // distancias no vlidas para celdas intermedias

    if (dist_if <= 1) then
       exit;

    // calculamos las alturas para la distancia dada
    ind:= 0;
    for i:= 0 to max-1 do begin

        for j:= 0 to repets[i]-1 do begin

            amVO[ind]:= (max-i) + celdaf.altura; // son alturas relativas al menor
            ind:= ind + 1;

        end;

    end;

    // liberamos la memoria del vector creado
    Finalize(repets);

end;


//------------------------------------------------------------------------------

(*
   Funcin que calcula el consumo de salud que implica el paso de una celda a otra en el mapa
   Recibe como parmetros:
      - celdai -> celda inicial del movimiento
      - celdaf -> celda final del movimiento
      - salud -> salud actual de la compaa
   Devuelve:
      - cantidad de salud que se consumira en el movimiento (en base a los tipos
        de las celdas y el impacto de armas)
*)
function consumo_salud (celdai: t_celda; celdaf: t_celda; salud: real):real;

begin

   // el dao sera la suma del dao que hace la celda
   // por defecto y el dao por impacto de arma en dicha celda
   Result:= celdaf.danio_0 + celdaf.danio_arma;   //(salud * celdaf.danio_arma);

end;

//------------------------------------------------------------------------------

(*
   Funcin que calcula el consumo de recursos que implica el paso de una celda a otra en el mapa
   Recibe como parmetros:
      - celdai -> celda inicial del movimiento
      - celdaf -> celda final del movimiento
   Devuelve:
      - cantidad de recursos que se consumiran en el movimiento (en base a los tipos
        de las celdas y las alturas respectivas)
*)
function consumo_recursos (celdai: t_celda; celdaf: t_celda; recursos: real):real;

var
   recur_cons, rec_alt: real;

begin

    // consideramos los recursos (por defecto) consumidos por atravesar la celda final
    recur_cons:= celdaf.recursos_0;

    // recursos consumidos por cambio de altura
    if (celdaf.altura >= celdai.altura) then
        rec_alt:= (celdaf.altura-celdai.altura)
    else
        rec_alt:= (celdai.altura-celdaf.altura)/2;      // si nos movemos hacia abajo, se
                                                        // consume la mitad de recursos

    // los recursos consumidos sern la suma de ambos
    Result:= recur_cons + rec_alt;

end;

//------------------------------------------------------------------------------

(*
   Funcin que calcula el consumo de recursos que implica el recorrido de una ruta
   (lista de celdas que constituyen un camino)
   Recibe como parmetros:
      - ruta -> lista de celdas que constituyen un camino
      - ini -> ndice del primer nodo a comprobar
      - fin -> ndice del ltimo nodo a mirar
   Devuelve:
      - cantidad de recursos que se consumiran en la ruta (en base a los tipos
        de las celdas y las alturas respectivas)
*)
function consumo_recursos_camino (ruta: t_camino; ini: integer; fin:integer): real;

var
  i: integer;

  nodo_act, nodo_sig: t_celda;  // nodos actual y siguiente en el camino
  recursos: real;

begin

   // el nodo inicial es el actual
   nodo_act:= ruta[ini];

   // inicializacin
   recursos:= 0;

   // recorremos los nodos del camino
   for i:= (ini+1) to fin do begin

       // nodo siguiente del camino
       nodo_sig:= ruta[i];

       // consideramos tambin el haberse movido por zonas ocultas (constante ponderacin)
       recursos:= recursos + consumo_recursos(nodo_act,nodo_sig,recursos);

       // el nuevo nodo actual es el siguiente de esta iteracin
       nodo_act:= nodo_sig;

   end;

   // resultado final
   Result:= recursos;

end;

//------------------------------------------------------------------------------

(*
   Funcin que calcula el consumo de salud que implica el recorrido de una ruta
   (lista de celdas que constituyen un camino)
   Recibe como parmetros:
      - ruta -> lista de celdas que constituyen un camino
      - ini -> ndice del primer nodo a comprobar
      - fin -> ndice del ltimo nodo a mirar
   Devuelve:
      - cantidad de salud que se consumira en la ruta (en base a los tipos
        de las celdas y al dao por impacto de arma)
*)
function consumo_salud_camino (ruta: t_camino; ini: integer; fin:integer): real;

var
  i: integer;

  nodo_act, nodo_sig: t_celda;  // nodos actual y siguiente en el camino
  salud: real;

begin

   // el nodo inicial es el actual
   nodo_act:= ruta[ini];

   // inicializacin
   salud:= 0;

   // recorremos los nodos del camino
   for i:= (ini+1) to fin do begin

       // nodo siguiente del camino
       nodo_sig:= ruta[i];

       // consideramos tambin el haberse movido por zonas ocultas (constante ponderacin)
       salud:= salud + consumo_salud(nodo_act,nodo_sig,salud);

       // el nuevo nodo actual es el siguiente de esta iteracin
       nodo_act:= nodo_sig;

   end;

   // resultado final
   Result:= salud;

end;

//------------------------------------------------------------------------------

(*
   Funcin que determina si una celda i es vista u oculta desde el punto de
   vista de otra f.
   Si las celdas que 'miran' son la unidad o un enemigo, considera para comprobar
   el rea de Capacidad de Adquisicin, que es la zona que tiene como radio dicho valor.
   Recibe como parmetros:
      - celdai -> celda a comprobar si es VO
      - celdaf -> celda desde la que se comprueba
      - zona -> tablero en el que estamos comprobando las zonas VO
      - max_x -> ndice mximo de x en la zona
      - max_y -> ndice mximo de y en la zona
   Devuelve:
      - 1 si es vista y 0 si es oculta
*)
function celda_VO (celdai: t_celda; celdaf: t_celda; var zona: t_tablero; max_x: integer; max_y: integer):byte;

var
  x,y: integer;   // coordenadas de las celdas a comprobar (entero)
  xr,yr: real;   // coordenadas de las celdas a comprobar (real)

  // incremento en x e y para seguir la recta DDA
  incx, incy: real;

  celda_c, celda_d: t_celda;  // celda a comprobar
  celda_k: t_celda;  // celda intermedia que se est comprobando

  dist_if, dis: integer;
  v_alt_max_VO: t_alt_max_VO;

  //dist_max: integer;  // distancia mxima a la que ve una celda a otra

  CeA: boolean;

begin

    // CONSIDERAMOS LA CAPACIDAD DE ADQUISICIN CORRESPONDIENTE
    if (celdaf.tipo_esp=ENEMIGO) then begin

       // si la celda a comprobar est fuera de la zona de adquisicin, ser oculta
       CeA:= celda_en_area_ca_hex(celdai,celdaf,CAP_ADQ_ENE,zona,max_x+1,max_y+1);
       if (CeA = FALSE) then begin
          Result:= 0;
          exit;
       end;

       //dist_max:= CAP_ADQ_ENE;
    end
    else if (celdaf.tipo_esp=PORIGEN) then begin

       // si la celda a comprobar est fuera de la zona de adquisicin, ser oculta
       CeA:= celda_en_area_ca_hex(celdai,celdaf,CAP_ADQ_UNI,zona,max_x+1,max_y+1);
       if (CeA = FALSE) then begin
          Result:= 0;
          exit;
       end;

       //dist_max:= CAP_ADQ_UNI;
    end;
    (*else
       dist_max:= High(integer);  // el mximo posible (no limita)*)

    // COMPROBACIONES INICIALES

    // si la celda a comprobar es de tipo bosque => ZONA OCULTA
    if (celdai.tipo = BOSQ) then begin
       Result:= 0;
       exit;
    end;

    // COMPROBACIONES DE LA RECTA QUE UNE A LAS CELDAS

    // consideraremos como celda inicial (o celda a comprobar) la de mayor altura

    // la celda con mayor altura ser la celda a comprobar ,'c'
    // y la otra ser 'd'
    if (celdai.altura >= celdaf.altura) then begin
        celda_c:= celdai;
        celda_d:= celdaf;
    end
    else begin
        celda_c:= celdaf;
        celda_d:= celdai;
    end;

    // seguimos el algoritmo de la lnea DDA para movernos entre las 2 celdas (inicial y final)
    // calculamos la distancia entre la celda i y la f

    // (dentro calculamos los incrementos de x e y para la recta DDA)
    incx:= 0;   incy:= 0;
    dist_if:= dist_DDA(celda_c,celda_d,incx,incy);

    (*// si las celdas distan ms de la distancia mxima (limitada a la capacidad
    // de adquisicin si la que 'mira' es el enemigo o la unidad), i SER OCULTA A f.
    if (dist_if>dist_max) then begin
       Result:= 0;
       exit;
    end;*)

    // si las celdas estn a distancia 1 siempre se vern
    // (salvo si la inicial es bosque, que ya se comprob)
    if (dist_if = 1) then begin
       Result:= 1;
       exit;
    end;


    // solo se calcula el vector de alturas si tienen alturas diferentes

    if (celdai.altura <> celdaf.altura) then begin

        // (*)  Creamos Vector de Alturas (*)

        // reservamos memoria para el vector de alturas
        SetLength(v_alt_max_VO,(dist_if-1));

        // creamos el vector de alturas
        calculo_alt_max_VO(celda_c,celda_d,dist_if,v_alt_max_VO);

    end;

    dis:= 0;

    // celdas iniciales
    // pasamos a la siguiente celda siguiendo la lnea con algoritmo DDA
    xr:= celda_c.x + incx;
    yr:= celda_c.y + incy;
    x:= round(xr);
    y:= round(yr);

    // recorremos la recta comprobando las condiciones de visibilidad
    while ((x <> celda_d.x) or (y <> celda_d.y)) and
          ((x <= max_x) and (y <= max_y)) do begin

        // guardamos los datos de la celda intermedia a comprobar
        celda_k:= zona[y,x];  // y=fila, x=columna

        // si la altura de la celda c es igual que la de la celda d
        if (celda_c.altura = celda_d.altura) then begin

            // si hay en medio una celda con la misma o mayor altura y
            // es de tipo bosque => ZONA OCULTA
            if (celda_k.altura >= celda_c.altura) and (celda_k.tipo = BOSQ) then begin
                Result:= 0;
                exit;
            end;

            // si hay una celda en medio con altura mayor en 2 o ms unidades
            // que la celda final => ZONA OCULTA
            if (celda_k.altura - celda_d.altura) > 1 then begin
                Result:= 0;
                exit;
            end;

        end
        // si las celdas tienen alturas diferentes
        else begin

            // si la altura de la celda es igual o mayor que la mxima permitida y es bosque => ZONA OCULTA
            if (celda_k.altura >= v_alt_max_VO[dis]) and (celda_k.tipo = BOSQ) then begin
                Result:= 0;
                exit;
            end;

            // si la celda tiene mayor altura que la mxima permitida => ZONA OCULTA
            if (celda_k.altura > v_alt_max_VO[dis]) then begin
                Result:= 0;
                exit;
            end;

        end;

        // pasamos a la siguiente celda siguiendo la lnea con algoritmo DDA
        xr:= xr + incx;
        yr:= yr + incy;

        x:= round(xr);
        y:= round(yr);

        dis:= dis + 1;

    end;  // while


    // liberar memoria al final
    Finalize(v_alt_max_VO);

    // si no se ha terminado an, es porque la celda es visible
    Result:= 1;

end;

//------------------------------------------------------------------------------

(*
   Funcin que determina si es posible pasar de una celda a otra.
   Se mira si la celda destino est definida como un obstculo (artificial)
   Se considera que hay un obstculo natural si la diferencia de alturas
   (en cualquier sentido) entre ambas celdas es > de MAX_DIF_ALTURAS_UNI.
   Recibe como parmetros:
      - celdai -> celda inicial del movimiento
      - celdaf -> celda final del movimiento
   Devuelve:
      - TRUE si hay un obstculo y FALSE si se puede pasar
*)
function hay_obstaculo(celdai: t_celda; celdaf: t_celda): boolean;
begin

  // si la celda destino es de tipo obstculo, hay un obstculo artificial
  if (celdaf.tipo = OBST) then begin
     Result:= TRUE;
     exit;
  end;

  // se calcula la diferencia de altura entre ambas y si es mayor de la
  // posible, hay un obstculo natural
  if (abs(celdai.altura - celdaf.altura) > MAX_DIF_ALTURAS_UNI) then
     Result:= TRUE
  else
     Result:= FALSE;

end;

//------------------------------------------------------------------------------

(*
   Funcin que busca a los enemigos dentro del rea hexagonal que determina
   un radio igual a la capacidad de adquisicin de la unidad.
   Recibe como parmetros:
      - celdau -> celda de la unidad
      - radio_ca -> radio de la zona hexagonal
      - Enems -> array de enemigos que se rellena con los que se encuentre
      - Tablero -> mapa en el que se comprueba
      - max_x -> lmite de columnas del mapa
      - max_y -> lmite de filas del mapa
   Devuelve:
      - el nmero de enemigos encontrados
*)

function buscar_enemigos_ca_hex(celdau: t_celda; radio_ca: integer; var Enems: t_lista_enemigos;
                               var Tablero: t_tablero; max_x: integer; max_y: integer): integer;
var
   i,j,fil,col,pos,c: integer;

   limNO_x, limNO_y: integer;
   inc_nc, num_cels: integer;

   celdac: t_celda;

   enems_encontrados: integer;

begin

  // calculamos las coordenadas del punto extremo NorOeste del rea
  // columna (x)
  limNO_x:= celdau.x - radio_ca;

  // si columna final es PAR
  if ((limNO_x mod 2) = 0) then begin
     // fila (y)
     limNO_y:= celdau.y - (radio_ca div 2);
  end
  else begin
     // Columna origen PAR
     if ((celdau.x mod 2) = 0) then
        // fila (y)
        limNO_y:= celdau.y - ((radio_ca div 2)+1)
     else
        // fila (y)
        limNO_y:= celdau.y - (radio_ca div 2);
  end;


  (** se recorren los nodos del rea de izquierda a derecha y de arriba a abajo **)

  // inicializaciones
  fil:= limNO_y;
  col:= limNO_x;

  inc_nc:= 1;
  num_cels:= radio_ca;
  enems_encontrados:= 0;
  
  if ((col mod 2)=0) then c:= 1
  else c:= 0;

  (* bucle exterior (de izquierda a derecha). De NorOeste a NorEste *)
  for i:= 1 to ((radio_ca*2)+1) do begin

     // nmero de celdas a explorar (de arriba a abajo)
     num_cels:= num_cels + inc_nc;

     (* bucle interior (de arriba a abajo) *)
     for j:=1 to num_cels do begin

        // si se sale de los lmites, no se hace nada
        if ((fil<0) or (fil>=max_y)) or ((col<0) or (col>=max_x)) then begin
           fil:= fil+1;
           continue;
        end;

        celdac:= Tablero[fil,col];

        // si la celda es la misma que la central, no se hace nada
        if (celdac.x = celdau.x) and (celdac.y = celdau.y) then begin
           fil:= fil+1;
           continue;
        end;

        // comprobar si es un enemigo
        if (celdac.tipo_esp = ENEMIGO) then begin
           // si es visible desde la posicin de la unidad, se apunta al enemigo
           // como visto
           if (celda_VO(celdac,celdau,Tablero,max_x,max_y) = 1) then begin
              // se aade el enemigo al array de enemigos
              pos:= High(Enems);
              SetLength(Enems,pos+2);
              Enems[pos+1]:= celdac;
              enems_encontrados:= enems_encontrados+1;
           end;

        end;

        // la fila va cambiando (creciendo hacia abajo)
        fil:= fil+1;

     end; // for j


     // se deja la fila como estaba
     fil:= fil-num_cels;

     c:= c+1;

     // columnas a la izquierda de la central
     if (i<radio_ca) then begin
        if (c=2) then begin
          c:= 0;
          fil:= fil-1;  // cada 2 iteraciones se resta 1
        end;
        col:= col+1;
     end
     // columna central
     else if (i=radio_ca) then begin
        if (c=2) then begin
          fil:= fil-1;  // cada 2 iteraciones se resta 1
        end;

        col:= col+1;

        // se inicializa c en funcin de si la celda est en columna PAR o IMPAR
        if ((col mod 2) = 0) then c:= 0
        else c:= 1;

        //fil:= fil+1;
     end
     // columnas a la derecha de la central
     else begin
        if (c=2) then begin
          c:= 0;
          fil:= fil+1;  // cada 2 iteraciones se suma 1
        end;
        col:= col+1;

        inc_nc:= -1;
     end;


  end; // for i

  Result:= enems_encontrados;

end;

//------------------------------------------------------------------------------

(*
   Funcin que busca a los enemigos dentro del rea que determinan las celdas
   que estn a una distancia igual a un radio dado (capacidad de adquisicin de la unidad).
   La distancia se calcula segn una lnea dibujada con el algoritmo DDA o
   la Distancia Eucldea (segn parmetro).
   Recibe como parmetros:
      - celdau -> celda de la unidad
      - radio_ca -> radio de la zona hexagonal
      - Enems -> array de enemigos que se rellena con los que se encuentre
      - Tablero -> mapa en el que se comprueba
      - max_x -> lmite de columnas del mapa
      - max_y -> lmite de filas del mapa
      - cad_dist -> indica el tipo de distancia a calcular ('DDA' o 'EUCLID')
   Devuelve:
      - el nmero de enemigos encontrados
*)

function buscar_enemigos_ca_dist(celdau: t_celda; radio_ca: integer; var Enems: t_lista_enemigos;
                                 var Tablero: t_tablero; max_x: integer; max_y: integer; cad_dist: string): integer;
var
   i,j,pos: integer;

   dist: real;
   incx,incy: real;

   celdac: t_celda;

   enems_encontrados: integer;

begin

  enems_encontrados:= 0;

  (** se recorren todos los nodos del mapa **)

  for i:= 0 to (max_y-1) do begin

     for j:=0 to (max_x-1) do begin

        celdac:= Tablero[i,j];

        // si la celda es la misma que la central, no se hace nada
        if (celdac.x = celdau.x) and (celdac.y = celdau.y) then
           continue;

        // se calcula la distancia elegida (DDA o Eucldea)
        if (cad_dist='DDA') then begin
          incx:= 0;  incy:= 0;
          dist:= dist_DDA(celdau,celdac,incx,incy);
        end
        else
          dist:= sqrt( sqr(celdac.x-celdau.x) + sqr(celdac.y-celdau.y) );

        // si la distancia es ms de la capac. de adquisicin, no estar dentro
        // del rea y se pasa a la siguiente
        if (dist>CAP_ADQ_UNI) then continue;

        // comprobar si es un enemigo
        if (celdac.tipo_esp = ENEMIGO) then begin
           // si es visible desde la posicin de la unidad, se apunta al enemigo
           // como visto
           if (celda_VO(celdac,celdau,Tablero,max_x,max_y) = 1) then begin
              // se aade el enemigo al array de enemigos
              pos:= High(Enems);
              SetLength(Enems,pos+2);
              Enems[pos+1]:= celdac;
              enems_encontrados:= enems_encontrados+1;
           end;

        end;

     end; // for j

  end; // for i

  Result:= enems_encontrados;

end;

//------------------------------------------------------------------------------

(*
   Funcin que comprueba si una celda est dentro del rea hexagonal que determina
   un radio (generalmente igual a la capacidad de adquisicin de la unidad o del enemigo).
   Recibe como parmetros:
      - celdac -> celda a comprobar si est en el rea
      - celdau -> celda central del rea (unidad o enemigo generalmente)
      - radio_ca -> radio de la zona hexagonal
      - Tablero -> mapa en el que se comprueba
      - max_x -> lmite de columnas del mapa
      - max_y -> lmite de filas del mapa
   Devuelve:
      - TRUE si la celdac pertenece al rea hexagonal de radio radio_ca de celdau
      - FALSE si est fuera de dicha rea
*)

function celda_en_area_ca_hex(celdac: t_celda; celdau: t_celda; radio_ca: integer;
                              var Tablero: t_tablero; max_x: integer; max_y: integer): boolean;
var
   i,j,fil,col,c: integer;

   limNO_x, limNO_y: integer;
   inc_nc, num_cels: integer;

   dist: real;

begin

  (* calculamos las coordenadas del punto extremo NorOeste del rea *)

  // columna (x)
  limNO_x:= celdau.x - radio_ca;

  // si columna final es PAR
  if ((limNO_x mod 2) = 0) then begin
     // fila (y)
     limNO_y:= celdau.y - (radio_ca div 2);
  end
  else begin
     // Columna origen PAR
     if ((celdau.x mod 2) = 0) then
        // fila (y)
        limNO_y:= celdau.y - ((radio_ca div 2)+1)
     else
        // fila (y)
        limNO_y:= celdau.y - (radio_ca div 2);
  end;


  (* Comprobaciones Rpidas *)

  // EST

  // si es la celda central
  if (celdac.x = celdau.x) and (celdac.y = celdau.y) then begin
     Result:= TRUE;
     exit;
  end;

  // NO EST


  // si la (distancia Eucldea - 1) al centro es mayor que el radio, estar fuera
  dist:= sqrt( sqr(celdac.x-celdau.x) + sqr(celdac.y-celdau.y) ) - 1;
  if (dist > (radio_ca+1)) then begin
     Result:= FALSE;
     exit;
  end;


  (* Comprobacin del rea *)
  (** se recorren los nodos del rea de izquierda a derecha y de arriba a abajo **)

  // inicializaciones
  fil:= limNO_y;
  col:= limNO_x;

  inc_nc:= 1;
  num_cels:= radio_ca;

  if ((col mod 2)=0) then c:= 1
  else c:= 0;

  (* bucle exterior (de izquierda a derecha). De NorOeste a NorEste *)
  for i:= 1 to ((radio_ca*2)+1) do begin

     // nmero de celdas a explorar (de arriba a abajo)
     num_cels:= num_cels + inc_nc;

     (* bucle interior (de arriba a abajo) *)
     for j:=1 to num_cels do begin

        // si se sale de los lmites, no se hace nada
        if ((fil<0) or (fil>=max_y)) or ((col<0) or (col>=max_x)) then begin
           fil:= fil+1;
           continue;
        end;

        // si la celda es la que buscamos, terminamos con TRUE
        if (celdac.x = Tablero[fil,col].x) and (celdac.y = Tablero[fil,col].y) then begin
           Result:=TRUE;
           exit;
        end;

        // la fila va cambiando (creciendo hacia abajo)
        fil:= fil+1;

     end; // for j


     // se deja la fila como estaba
     fil:= fil-num_cels;

     c:= c+1;

     // columnas a la izquierda de la central
     if (i<radio_ca) then begin
        if (c=2) then begin
          c:= 0;
          fil:= fil-1;  // cada 2 iteraciones se resta 1
        end;
        col:= col+1;
     end
     // columna central
     else if (i=radio_ca) then begin
        if (c=2) then begin
          fil:= fil-1;  // cada 2 iteraciones se resta 1
        end;

        col:= col+1;

        // se inicializa c en funcin de si la celda est en columna PAR o IMPAR
        if ((col mod 2) = 0) then c:= 0
        else c:= 1;

        //fil:= fil+1;
     end
     // columnas a la derecha de la central
     else begin
        if (c=2) then begin
          c:= 0;
          fil:= fil+1;  // cada 2 iteraciones se suma 1
        end;
        col:= col+1;

        inc_nc:= -1;
     end;


  end; // for i

  // si se llega aqu, es porque la celda no est
  Result:= FALSE;

end;

end.
