(* Tipos y funciones usadas por los algoritmos implementados *)

unit Tipos;

interface

  uses Simulador;

(* ********************* TIPOS DEFINIDOS ********************** *)
type

  // tipo nodos alcanzables (necesario para pasar array dinmico a funcin)
  t_alcanzables = array of t_celda;

  // estructura con la Solucin (fitness de cada objetivo, consumos de recursos y salud,
  // y camino solucin)
  t_solucion = record
    fit_obj1, fit_obj2: real;
    consum_rec, consum_sal: real;
    camino: t_camino;
  end;

  // estructura para los datos de las hormigas
  t_hormiga = record
     salud: real;
     recursos: real;
     solucion: t_solucion;
  end;

  // estructura para guardar el Frente de Pareto
  t_frente_pareto = record
     num_sols_FP: integer;                  // nmero de solucs que hay en el FP
     soluciones_FP: array of t_solucion;    // soluciones del FP
  end;

  // estructura para guardar los datos estadsticos del FP
  t_estadisticas_FP = record
     min_fit1, min_fit2: real;               // fitness mnimo
     fit_obj1_r, fit_obj2_r: real;           // fitness minimo de los objetivos para camino ms rpido
     fit_obj1_s, fit_obj2_s: real;           // fitness minimo de los objetivos para camino ms seguro
     med_fit_obj1, med_fit_obj2: real;   // fitness medio de los objetivos
  end;

(* ********************* VARIABLES GLOBALES ********************** *)
var
  // --------------------- Constantes del Programa ----------------------

  // Parmetros del SCH
  ALPHA: byte;    // ponderacin feromona (1)
  BETA: byte;     // ponderacin heurstica (2)
  RHO: single;     // parmetro de evaporacin de feromona (0.1)
  q0: single;     // parmetro de explotacin/exploracin en SCH (0.4)

  PHI: single;    // parmetro de evaporacin local de feromona en SCH monoobjetivo

  // Constantes de Ponderacin de las Heursticas
  // N1 (Camino ms Rpido)
  KN1_REC: single;    // el consumo de recursos es importante (0.1)
  KN1_VO:  single;    // las zonas vistas y ocultas es secundario (0.05)
  KN1_DIS: single;    // la distancia debe irse reduciendo (0.85)

  // N2 (Camino ms Seguro)
  KN2_DAN: single;    // el evitar daos es importante (0.1)
  KN2_VO:  single;    // las zonas vistas y ocultas es muy importante (0.85)
  KN2_DIS: single;    // la distancia al punto final (poco relevante, pero que hay que considerar) (0.05)

  // Constantes de Ponderacin de la Funcin de Evaluacin
  KEV1_ZVO: single;   // Zonas Vistas/Ocultas en objetivo 1: camino ms rpido (0.5)
  KEV2_ZVO: single;   // Zonas Vistas/Ocultas en objetivo 2: camino ms seguro (10)


(* **************** FUNCIONES Y PROCEDIMIENTOS PBLICOS ***************** *)

function distancia_eu (nodoi: t_celda; nodof: t_celda):real;
function es_vecino_hexa (fila: integer; columna: integer;
                         fila_v: integer; columna_v: integer;
                         N_FIL_TAB, N_COL_TAB: integer): boolean;

(* **************** IMPLEMENTACIN DE FUNCIONES Y PROCEDIMIENTOS ***************** *)

implementation



(*
   Funcin que calcula la distancia eucldea entre 2 celdas
   Recibe como parmetros:
      - nodoi -> nodo inicial
      - nodof -> nodo final
   Devuelve:
      - distancia eucldea entre los nodos
*)
function distancia_eu (nodoi: t_celda; nodof: t_celda):real;
begin
    Result:= sqrt( sqr(nodof.x-nodoi.x) + sqr(nodof.y-nodoi.y) );
end;

//------------------------------------------------------------------------------

(* @@@
   Funcin que comprueba si una celda pertenece al vecindario de otra en un
   grid hexagonal.
   Recibe como parmetros:
      - fila -> fila del nodo actual
      - columna -> columna del nodo actual
      - fila_v -> fila del vecino
      - columna_v -> columna del vecino
      - N_FIL_TAB -> filas totales del tablero
      - N_COL_TAB -> columnas totales del tablero
   Devuelve:
      - TRUE si el nodo es vecino en un grid hexagonal y FALSE en caso contrario
*)
function es_vecino_hexa (fila: integer; columna: integer;
                         fila_v: integer; columna_v: integer;
                         N_FIL_TAB, N_COL_TAB: integer): boolean;
begin

   // @@@ Si se sale de alguno de los lmites del tablero no sern vecinos
  if (fila_v < 0) or (fila_v >= N_FIL_TAB) then begin
     Result:= FALSE;
     exit;
  end;

  // @@@ si se sale de los lmites del tablero, ya que no ser un nodo a atravesar
  if (columna_v < 0) or (columna_v >= N_COL_TAB) then begin
     Result:= FALSE;
     exit;
  end;

  // @@@ si distan ms de 1 en algn sentido, no sern vecinos
  if (abs(fila-fila_v)>1) or (abs(columna-columna_v)>1) then begin
     Result:= FALSE;
     exit;
  end;

  Result:= TRUE;

  // @@@ La celda no ser vecino en el hexgono si:
  if ((columna mod 2) = 0) then begin
     // @@@ La columna de la central es PAR y las coordenadas de la posible vecina son
     // @@@ (fil+1,col-1) o (fil+1,col+1)
     if ((fila_v = (fila+1)) and (columna_v = (columna-1))) or
        ((fila_v = (fila+1)) and (columna_v = (columna+1))) then begin
        Result:= FALSE;
     end;
  end
  else begin
     // @@@ o Su columna es IMPAR y las coordenadas de la posible vecina son
     // @@@ (fil-1,col-1) o (fil-1,col+1)
     if ((fila_v = (fila-1)) and (columna_v = (columna-1))) or
        ((fila_v = (fila-1)) and (columna_v = (columna+1))) then begin
        Result:= FALSE;
     end;
  end;

end;


end.
